"""Deprecated module for Web API resource classes and registration.

.. deprecated:: 0.9

See the following for the new function/class locations:

* :py:class:`djblets.webapi.resources.base.WebAPIResource`
* :py:class:`djblets.webapi.resources.group.GroupResource`
* :py:data:`djblets.webapi.resources.group.group_resource`
* :py:class:`djblets.webapi.resources.root.RootResource`
* :py:class:`djblets.webapi.resources.user.UserResource`
* :py:data:`djblets.webapi.resources.user.user_resource`
* :py:func:`djblets.webapi.resources.registry.get_resource_from_class`
* :py:func:`djblets.webapi.resources.registry.get_resource_from_name`
* :py:func:`djblets.webapi.resources.registry.get_resource_for_object`
* :py:func:`djblets.webapi.resources.registry.register_resource_for_model`
* :py:func:`djblets.webapi.resources.registry.unregister_resource`
* :py:func:`djblets.webapi.resources.registry.unregister_resource_for_model`
"""

from __future__ import unicode_literals

from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.resources.group import GroupResource, group_resource
from djblets.webapi.resources.registry import (get_resource_from_class,
                                               get_resource_from_name,
                                               get_resource_for_object,
                                               register_resource_for_model,
                                               unregister_resource,
                                               unregister_resource_for_model)
from djblets.webapi.resources.root import RootResource
from djblets.webapi.resources.user import UserResource, user_resource


__all__ = [
    'GroupResource',
    'RootResource',
    'UserResource',
    'WebAPIResource',
    'get_resource_from_class',
    'get_resource_from_name',
    'get_resource_for_object',
    'register_resource_for_model',
    'unregister_resource',
    'unregister_resource_for_model',
    'group_resource',
    'user_resource',
]

__deprecated__ = __all__



"""Unit tests for djblets.privacy.consent.base.BaseConsentRequirement."""

from __future__ import unicode_literals

from datetime import datetime

from django.contrib.auth.models import User
from django.utils import timezone

from djblets.privacy.consent import (BaseConsentRequirement, Consent,
                                     ConsentData, get_consent_tracker)
from djblets.privacy.tests.testcases import ConsentTestCase


class MyConsentRequirement(BaseConsentRequirement):
    requirement_id = 'my-requirement'
    name = 'My Requirement'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'


class BaseConsentRequirementTests(ConsentTestCase):
    """Unit tests for djblets.privacy.consent.base.BaseConsentRequirement."""

    def test_build_consent_data(self):
        """Testing BaseConsentRequirement.build_consent_data"""
        requirement = MyConsentRequirement()
        timestamp = datetime(2018, 1, 2, 13, 14, 15, tzinfo=timezone.utc)

        consent_data = requirement.build_consent_data(
            granted=False,
            timestamp=timestamp,
            source='http://example.com/account/profile/#consent',
            extra_data={
                'test': True,
            })

        self.assertEqual(consent_data.requirement_id, 'my-requirement')
        self.assertFalse(consent_data.granted)
        self.assertEqual(consent_data.timestamp, timestamp)
        self.assertEqual(consent_data.source,
                         'http://example.com/account/profile/#consent')
        self.assertEqual(
            consent_data.extra_data,
            {
                'test': True,
            })

    def test_get_consent(self):
        """Testing BaseConsentRequirement.get_consent"""
        requirement = MyConsentRequirement()
        timestamp = datetime(2018, 1, 2, 13, 14, 15, tzinfo=timezone.utc)
        user = User.objects.create(username='test-user')

        consent_data = ConsentData(
            requirement_id='my-requirement',
            granted=True,
            timestamp=timestamp,
            source='http://example.com/account/profile/#consent',
            extra_data={
                'test': True,
            })

        get_consent_tracker().record_consent_data(user, consent_data)

        self.assertEqual(requirement.get_consent(user), Consent.GRANTED)



from __future__ import unicode_literals

import json

from djblets.testing.testcases import TestCase
from djblets.webapi.encoders import (JSONEncoderAdapter, WebAPIEncoder,
                                     XMLEncoderAdapter)


class EncoderAdapterTests(TestCase):
    """Tests encoding correctness of WebAPIEncoder adapters"""
    data = {
        'dict_val': {'foo': 'bar'},
        'list_val': [10, 'baz'],
        'string_val': 'foobar',
        'int_val': 42,
        'float_val': 3.14159,
        'scientific_val': 2.75e-15,
        'bool_val': True,
        "none_val": None
    }

    def test_json_encoder_adapter(self):
        """Testing JSONEncoderAdapter.encode"""
        encoder = WebAPIEncoder()
        adapter = JSONEncoderAdapter(encoder)

        content = adapter.encode(self.data)
        self.assertEqual(content, json.dumps(self.data))

    def test_xml_encoder_adapter(self):
        """Testing XMLEncoderAdapter.encode"""
        encoder = WebAPIEncoder()
        adapter = XMLEncoderAdapter(encoder)

        expected = (
            '<?xml version="1.0" encoding="utf-8"?>\n'
            '<rsp>\n'
            ' <string_val>foobar</string_val>\n'
            ' <none_val>\n'
            ' </none_val>\n'
            ' <dict_val>\n'
            '  <foo>bar</foo>\n'
            ' </dict_val>\n'
            ' <bool_val>1</bool_val>\n'
            ' <scientific_val>2.75e-15</scientific_val>\n'
            ' <int_val>42</int_val>\n'
            ' <float_val>3.14159</float_val>\n'
            ' <list_val>\n'
            '  <array>\n'
            '   <item>10</item>\n'
            '   <item>baz</item>\n'
            '  </array>\n'
            ' </list_val>\n'
            '</rsp>'
        )

        content = adapter.encode(self.data)
        self.assertEqual(content, expected)



"""Version information for certain Djblets dependencies.

This contains constants that other parts of Djblets and consumers of Djblets
can use to look up information on major dependencies of Djblets.

The contents in this file might change substantially between releases. If
you're going to make use of data from this file, code defensively.
"""

from __future__ import unicode_literals

# NOTE: This file may not import other files! It's used for packaging and
#       may be needed before any dependencies have been installed.


#: The major version of Django we're using for documentation.
django_doc_major_version = '1.6'

#: The version range required for Django.
django_version = '>=1.6.11,<1.10.999'

#: Dependencies required for LessCSS pipelining.
lesscss_npm_dependencies = {
    'less': '2.6.0',
    'less-plugin-autoprefix': '1.5.1',
}

#: Dependencies required for UglifyJS JavaScript compression.
uglifyjs_npm_dependencies = {
    'uglify-js': '2.4.10',
}

#: Dependencies required for Babel for JavaScript.
babel_npm_dependencies = {
    'babel-cli': '6.5.1',
    'babel-preset-es2015': '6.5.0',
    'babel-plugin-dedent': '2.0.0',
}

#: All static media dependencies required to package/develop against  Djblets.
npm_dependencies = {}
npm_dependencies.update(lesscss_npm_dependencies)
npm_dependencies.update(uglifyjs_npm_dependencies)
npm_dependencies.update(babel_npm_dependencies)

#: All dependencies required to install Djblets.
package_dependencies = {
    'Django': django_version,
    'django-pipeline': '>=1.6.14,<1.6.999',
    'dnspython': '>=1.14.0',
    'feedparser': '>=5.1.2',
    'pillowfight': '',
    'publicsuffix': '>=1.1',
    'python-dateutil': '>=1.5',
    'pytz': '',
}


def build_dependency_list(deps, version_prefix=''):
    """Build a list of dependency specifiers from a dependency map.

    This can be used along with :py:data:`package_dependencies`,
    :py:data:`npm_dependencies`, or other dependency dictionaries to build a
    list of dependency specifiers for use on the command line or in
    :file:`setup.py`.

    Args:
        deps (dict):
            A dictionary of dependencies.

    Returns:
        list of unicode:
        A list of dependency specifiers.
    """
    return [
        '%s%s%s' % (dep_name, version_prefix, dep_version)
        for dep_name, dep_version in deps.items()
    ]



"""Utility mixins for configuration forms."""

from __future__ import unicode_literals


class DynamicConfigPageMixin(object):
    """A mixin for utilizing a registry to keep track of the pages and forms.

    Objects using this mixin can have a dynamic list of forms, which can be
    updated by callers. New forms can be registered on them and existing forms
    can be removed. With this mixin, it is possible to extend a configuration
    page, allowing each form to load, validate, and save without the page
    hard-coding support for each form.

    Forms utilizing this mixin are required to set the :py:attr:`registry`
    attribute to an instance of
    :py:class:`~djblets.configforms.registry.ConfigPageRegistry`.

    To manage the forms on a page, the :py:meth:`add_form` and
    :py:meth:`remove_form` methods should be used.
    """

    #: The registry to use.
    #:
    #: This should be an instance of
    # :py:class:`~djblets.configforms.registry.ConfigPageRegistry`.
    registry = None

    _default_form_classes = None

    @classmethod
    def add_form(cls, form_class):
        """Add a form to the page.

        Args:
            form_class (type):
                The form to add to the page, as a subclass of
                :py:class:`~djblets.configforms.page.ConfigPage`.
        """
        cls.registry.add_form_to_page(cls, form_class)

    @classmethod
    def remove_form(cls, form_class):
        """Remove a form from the page.

        Args:
            form_class (type):
                The form to remove from the page, as a subclass of
                :py:class:`~djblets.configforms.forms.ConfigPageForm`.
        """
        cls.registry.remove_form_from_page(cls, form_class)



"""Unit tests for PrePostDispatchViewMixin."""

from __future__ import unicode_literals

from django.http import HttpResponse, HttpResponseForbidden
from django.test.client import RequestFactory
from django.views.generic.base import View

from djblets.testing.testcases import TestCase
from djblets.views.generic.base import PrePostDispatchViewMixin


class PrePostDispatchViewMixinTests(TestCase):
    """Unit tests for PrePostDispatchViewMixin."""

    def test_dispatch(self):
        """Testing PrePostDispatchViewMixin.dispatch"""
        seen = set()

        class MyView(PrePostDispatchViewMixin, View):
            def pre_dispatch(self, *args, **kwargs):
                seen.add('pre_dispatch')

            def post_dispatch(self, response, *args, **kwargs):
                seen.add('post_dispatch')

                return response

        view = MyView.as_view()
        response = view(RequestFactory().request())

        self.assertEquals(seen, {'pre_dispatch', 'post_dispatch'})
        self.assertIsInstance(response, HttpResponse)

    def test_dispatch_with_pre_dispatch_response(self):
        """Testing PrePostDispatchViewMixin.pre_dispatch with custom response
        """
        class MyView(PrePostDispatchViewMixin, View):
            def pre_dispatch(self, *args, **kwargs):
                return HttpResponseForbidden()

        view = MyView.as_view()
        response = view(RequestFactory().request())

        self.assertIsInstance(response, HttpResponseForbidden)

    def test_dispatch_with_post_dispatch_response(self):
        """Testing PrePostDispatchViewMixin.post_dispatch with custom response
        """
        class MyView(PrePostDispatchViewMixin, View):
            def post_dispatch(self, *args, **kwargs):
                return HttpResponseForbidden()

        view = MyView.as_view()
        response = view(RequestFactory().request())

        self.assertIsInstance(response, HttpResponseForbidden)



"""Unit tests for djblets.privacy.consent.tracker."""

from __future__ import unicode_literals

from datetime import datetime

from django.contrib.auth.models import User
from django.core.cache import cache
from django.test.utils import override_settings
from django.utils import timezone
from kgb import SpyAgency

from djblets.cache.backend import make_cache_key
from djblets.privacy.consent import (BaseConsentRequirement,
                                     BaseConsentTracker, Consent, ConsentData,
                                     DatabaseConsentTracker,
                                     get_consent_requirements_registry,
                                     get_consent_tracker)
from djblets.privacy.models import StoredConsentData
from djblets.privacy.tests.testcases import ConsentTestCase


class CustomConsentTracker(BaseConsentTracker):
    pass


class MyConsentRequirement1(BaseConsentRequirement):
    requirement_id = 'my-requirement-1'
    name = 'My Requirement 1'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'
    icons = {
        '1x': '/static/consent.png',
        '2x': '/static/consent@2x.png',
    }


class MyConsentRequirement2(BaseConsentRequirement):
    requirement_id = 'my-requirement-2'
    name = 'My Requirement 2'
    summary = 'We would also like this'
    intent_description = 'We need this for dancing.'
    data_use_description = 'Dancing all the things.'


class DatabaseConsentTrackerTests(SpyAgency, ConsentTestCase):
    """Unit tests for DatabaseConsentTracker."""

    def setUp(self):
        super(DatabaseConsentTrackerTests, self).setUp()

        self.tracker = DatabaseConsentTracker()
        self.user = User.objects.create(username='test-user',
                                        email='test@example.com')
        self.timestamp = datetime(2018, 1, 2, 13, 14, 15, tzinfo=timezone.utc)

        self.spy_on(timezone.now, call_fake=lambda: self.timestamp)

    def test_record_consent_data(self):
        """Testing DatabaseConsentTracker.record_consent_data"""
        consent_data = ConsentData(
            requirement_id='test-requirement-1',
            granted=True,
            source='http://example.com/account/profile/#consent',
            extra_data={
                'test': True,
            })

        self.tracker.record_consent_data(self.user, consent_data)

        stored_consents = list(StoredConsentData.objects.all())
        self.assertEqual(len(stored_consents), 1)

        stored_consent = stored_consents[0]
        self.assertEqual(stored_consent.user, self.user)
        self.assertEqual(stored_consent.audit_identifier,
                         '973dfe463ec85785f5f95af5ba3906eedb2d931c24e69'
                         '824a89ea65dba4e813b')
        self.assertEqual(stored_consent.time_added, self.timestamp)
        self.assertEqual(stored_consent.last_updated, self.timestamp)
        self.assertEqual(
            stored_consent.consent_grants,
            {
                'test-requirement-1': True,
            })
        self.assertEqual(
            stored_consent.audit_trail,
            {
                'test-requirement-1': [{
                    'identifier': '973dfe463ec85785f5f95af5ba3906eedb2d931c2'
                                  '4e69824a89ea65dba4e813b',
                    'granted': True,
                    'timestamp': '2018-01-02T13:14:15+00:00',
                    'source': 'http://example.com/account/profile/#consent',
                    'extra_data': {
                        'test': True,
                    },
                }],
            })

    def test_record_consent_data_list(self):
        """Testing DatabaseConsentTracker.record_consent_data_list"""
        consent_data_1 = ConsentData(
            requirement_id='test-requirement-1',
            granted=True,
            source='http://example.com/account/profile/#consent',
            extra_data={
                'test': True,
            })

        consent_data_2 = ConsentData(
            requirement_id='test-requirement-2',
            granted=False,
            source='http://example.com/account/profile/#consent')

        self.tracker.record_consent_data_list(self.user,
                                              [consent_data_1, consent_data_2])

        stored_consents = list(StoredConsentData.objects.all())
        self.assertEqual(len(stored_consents), 1)

        stored_consent = stored_consents[0]
        self.assertEqual(stored_consent.user, self.user)
        self.assertEqual(stored_consent.audit_identifier,
                         '973dfe463ec85785f5f95af5ba3906eedb2d931c24e69'
                         '824a89ea65dba4e813b')
        self.assertEqual(stored_consent.time_added, self.timestamp)
        self.assertEqual(stored_consent.last_updated, self.timestamp)
        self.assertEqual(
            stored_consent.consent_grants,
            {
                'test-requirement-1': True,
                'test-requirement-2': False,
            })
        self.assertEqual(
            stored_consent.audit_trail,
            {
                'test-requirement-1': [{
                    'identifier': '973dfe463ec85785f5f95af5ba3906eedb2d931c2'
                                  '4e69824a89ea65dba4e813b',
                    'granted': True,
                    'timestamp': '2018-01-02T13:14:15+00:00',
                    'source': 'http://example.com/account/profile/#consent',
                    'extra_data': {
                        'test': True,
                    },
                }],
                'test-requirement-2': [{
                    'identifier': '973dfe463ec85785f5f95af5ba3906eedb2d931c2'
                                  '4e69824a89ea65dba4e813b',
                    'granted': False,
                    'timestamp': '2018-01-02T13:14:15+00:00',
                    'source': 'http://example.com/account/profile/#consent',
                }],
            })

    def test_record_consent_data_list_clears_cache(self):
        """Testing DatabaseConsentTracker.record_consent_data_list clears cache
        """
        consent_data_1 = ConsentData(requirement_id='test-requirement-1',
                                     granted=True)
        consent_data_2 = ConsentData(requirement_id='test-requirement-2',
                                     granted=False)

        cache_key = make_cache_key('privacy-consent:%s' % self.user.pk)
        cache.add(cache_key, ['test-requirement-1'])

        self.assertEqual(cache.get(cache_key), ['test-requirement-1'])

        self.tracker.record_consent_data_list(self.user,
                                              [consent_data_1, consent_data_2])

        self.assertIsNone(cache.get(cache_key))

    def test_get_consent(self):
        """Testing DatabaseConsentTracker.get_consent"""
        # Populate some data we can fetch.
        self.test_record_consent_data_list()

        self.assertEqual(
            self.tracker.get_consent(self.user, 'test-requirement-1'),
            Consent.GRANTED)

        self.assertEqual(
            self.tracker.get_consent(self.user, 'test-requirement-2'),
            Consent.DENIED)

        self.assertEqual(
            self.tracker.get_consent(self.user, 'test-requirement-3'),
            Consent.UNSET)

        self.assertEqual(
            cache.get(make_cache_key('privacy-consent:%s' % self.user.pk)),
            {
                'test-requirement-1': Consent.GRANTED,
                'test-requirement-2': Consent.DENIED,
            })

    def test_get_consent_with_no_user_data(self):
        """Testing DatabaseConsentTracker.get_consent with user without any
        consent data
        """
        self.assertEqual(self.tracker.get_consent(self.user, 'test'),
                         Consent.UNSET)

        self.assertEqual(
            cache.get(make_cache_key('privacy-consent:%s' % self.user.pk)),
            {})

    def test_get_all_consent(self):
        """Testing DatabaseConsentTracker.get_all_consent"""
        # Populate some data we can fetch.
        self.test_record_consent_data_list()

        self.assertEqual(
            self.tracker.get_all_consent(self.user),
            {
                'test-requirement-1': Consent.GRANTED,
                'test-requirement-2': Consent.DENIED,
            })

        self.assertEqual(
            cache.get(make_cache_key('privacy-consent:%s' % self.user.pk)),
            {
                'test-requirement-1': Consent.GRANTED,
                'test-requirement-2': Consent.DENIED,
            })

    def test_get_all_consent_with_no_user_data(self):
        """Testing DatabaseConsentTracker.get_all_consent with user without any
        consent data
        """
        self.assertEqual(self.tracker.get_all_consent(self.user), {})

        self.assertEqual(
            cache.get(make_cache_key('privacy-consent:%s' % self.user.pk)),
            {})

    def test_get_pending_consent_requirements(self):
        """Testing DatabaseConsentTracker.get_pending_consent_requirements"""
        requirement1 = MyConsentRequirement1()
        requirement2 = MyConsentRequirement2()

        registry = get_consent_requirements_registry()

        try:
            registry.register(requirement1)
            registry.register(requirement2)

            self.assertEqual(
                self.tracker.get_pending_consent_requirements(self.user),
                [requirement1, requirement2])

            consent_data_1 = requirement1.build_consent_data(granted=True)
            self.tracker.record_consent_data_list(self.user, [consent_data_1])

            self.assertEqual(
                self.tracker.get_pending_consent_requirements(self.user),
                [requirement2])

            consent_data_2 = requirement2.build_consent_data(granted=True)
            self.tracker.record_consent_data_list(self.user, [consent_data_2])

            self.assertEqual(
                self.tracker.get_pending_consent_requirements(self.user),
                [])
        finally:
            registry.unregister(requirement1)
            registry.unregister(requirement2)


class ConsentTrackerInstanceTests(ConsentTestCase):
    """Unit tests for consent tracker instance management."""

    def test_get_consent_tracker_with_default(self):
        """Testing get_consent_tracker with default tracker"""
        self.assertIsInstance(get_consent_tracker(), DatabaseConsentTracker)

    @override_settings(
        DJBLETS_PRIVACY_CONSENT_TRACKER=('%s.CustomConsentTracker'
                                         % __name__))
    def test_get_consent_tracker_with_custom(self):
        """Testing get_consent_tracker with custom tracker"""
        self.assertIsInstance(get_consent_tracker(), CustomConsentTracker)



#
# djblets_email.py -- E-mail formatting template tags
#
# Copyright (c) 2007-2009  Christian Hammond
# Copyright (c) 2007-2009  David Trowbridge
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

import re

from django import template
from django.utils.safestring import mark_safe

from djblets.util.compat.django.template.loader import render_to_string
from djblets.util.decorators import blocktag


register = template.Library()


@register.simple_tag(takes_context=True)
def quoted_email(context, template_name):
    """
    Renders a specified template as a quoted reply, using the current context.
    """
    return quote_text(render_to_string(template_name, context))


@register.tag
@blocktag
def condense(context, nodelist, max_newlines=3):
    """Condenses a block of text.

    This will ensure that there are never more than the given number of
    consecutive newlines. It's particularly useful when formatting plain text
    output, to avoid issues with template tags adding unwanted newlines.
    """
    text = nodelist.render(context).strip()
    text = re.sub(r'\n{%d,}' % (max_newlines + 1), '\n' * max_newlines, text)
    return text


@register.filter
def quote_text(text, level=1):
    """
    Quotes a block of text the specified number of times.
    """
    lines = text.split("\n")
    quoted = ""

    for line in lines:
        quoted += "%s%s\n" % ("> " * level, line)

    return mark_safe(quoted.rstrip())



"""App configuration for djblets.privacy."""

from __future__ import unicode_literals

try:
    from django.apps import AppConfig
except ImportError:
    # Django < 1.7
    AppConfig = object


class PrivacyAppConfig(AppConfig):
    """Default app configuration for djblets.privacy."""

    name = 'djblets.privacy'
    label = 'djblets_privacy'



#
# managers.py -- Managers for Django database models.
#
# Copyright (c) 2007-2013  Beanbag, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import unicode_literals

from django.db import models, IntegrityError


class ConcurrencyManager(models.Manager):
    """
    A class designed to work around database concurrency issues.
    """
    def get_or_create(self, **kwargs):
        """
        A wrapper around get_or_create that makes a final attempt to get
        the object if the creation fails.

        This helps with race conditions in the database where, between the
        original get() and the create(), another process created the object,
        causing us to fail. We'll then execute a get().

        This is still prone to race conditions, but they're even more rare.
        A delete() would have to happen before the unexpected create() but
        before the get().
        """
        try:
            return super(ConcurrencyManager, self).get_or_create(**kwargs)
        except IntegrityError:
            kwargs.pop('defaults', None)
            return self.get(**kwargs)



from __future__ import unicode_literals


class FeatureLevel(object):
    """Possible stability levels for features."""

    #: The feature is completely unavailable.
    #:
    #: Feature checkers will not be consulted in this case.
    UNAVAILABLE = 0

    #: The feature is experimental.
    #:
    #: A feature checker must enable this feature.
    EXPERIMENTAL = 10

    #: The feature is in beta.
    #:
    #: A feature checker must enable this feature.
    BETA = 50

    #: The feature is stable and will always be enabled.
    STABLE = 100



"""A base form for working with settings stored on SiteConfigurations."""

from __future__ import unicode_literals

from djblets.forms.forms import KeyValueForm


class SiteSettingsForm(KeyValueForm):
    """A base form for loading/saving settings for a SiteConfiguration.

    This is meant to be subclassed for different settings pages. Any fields
    defined by the form will be loaded/saved automatically.

    Attributes:
        request (django.http.HttpRequest):
            The HTTP request used for this form.

        siteconfig (djblets.siteconfig.models.SiteConfiguration):
            The site configuration settings are loaded from and saved to.
    """

    def __init__(self, siteconfig, *args, **kwargs):
        """Initialize the form.

        Args:
            siteconfig (djblets.siteconfig.models.SiteConfiguration):
                The site configuration to work with on the form.

            *args (tuple):
                Positional arguments to pass to the parent constructor.

            **kwargs (dict):
                Keyword arguments to pass to the parent constructor.
        """
        self.request = kwargs.pop('request', None)
        self.siteconfig = siteconfig

        super(SiteSettingsForm, self).__init__(instance=siteconfig,
                                               *args, **kwargs)

    def get_key_value(self, key, default=None):
        """Return the value for a SiteConfiguration settings key.

        Args:
            key (unicode):
                The settings key.

            default (object):
                The default value from the form, which will be ignored,
                so that the registered siteconfig defaults will be used.

        Returns:
            object:
            The resulting value from the settings.
        """
        return self.instance.get(key)

    def set_key_value(self, key, value):
        """Set the value for a SiteConfiguration settings key.

        Args:
            key (unicode):
                The settings key.

            value (object):
                The settings value.
        """
        self.instance.set(key, value)

    def save_instance(self):
        """Save the SiteConfiguration instance."""
        self.instance.save()



from __future__ import unicode_literals

from djblets.testing.testcases import TestCase
from djblets.webapi.errors import WebAPIError


class WebAPIErrorTests(TestCase):
    """Unit tests for djblets.webapi.errors."""

    def test_with_message(self):
        """Testing WebAPIError.with_message"""
        orig_msg = 'Original message'
        new_msg = 'New message'
        headers = {
            'foo': 'bar',
        }

        orig_error = WebAPIError(123, orig_msg, http_status=500,
                                 headers=headers)
        new_error = orig_error.with_message(new_msg)

        self.assertNotEqual(orig_error, new_error)
        self.assertEqual(new_error.msg, new_msg)
        self.assertEqual(new_error.headers, headers)
        self.assertEqual(new_error.code, orig_error.code)
        self.assertEqual(new_error.http_status, orig_error.http_status)
        self.assertEqual(orig_error.msg, orig_msg)
        self.assertEqual(orig_error.headers, headers)

    def test_with_overrides(self):
        """Testing WebAPIError.with_overrides"""
        orig_msg = 'Original message'
        new_msg = 'New message'
        orig_headers = {
            'foo': 'bar',
        }
        new_headers = {
            'abc': '123',
        }

        orig_error = WebAPIError(123, orig_msg, http_status=500,
                                 headers=orig_headers)
        new_error = orig_error.with_overrides(new_msg, headers=new_headers)

        self.assertNotEqual(orig_error, new_error)
        self.assertEqual(new_error.msg, new_msg)
        self.assertEqual(new_error.headers, new_headers)
        self.assertEqual(new_error.code, orig_error.code)
        self.assertEqual(new_error.http_status, orig_error.http_status)
        self.assertEqual(orig_error.msg, orig_msg)
        self.assertEqual(orig_error.headers, orig_headers)



"""Error classes for user consents."""

from __future__ import unicode_literals

from djblets.registries.errors import AlreadyRegisteredError, ItemLookupError


class ConsentRequirementConflictError(AlreadyRegisteredError):
    """A consent requirement with a given ID was already registered."""


class ConsentRequirementNotFoundError(ItemLookupError):
    """A consent requirement was not found."""



"""Unit tests for djblets.privacy.templatetags.djblets_privacy."""

from __future__ import unicode_literals

from django.template import Context, Template
from django.test.client import RequestFactory

try:
    # Django >= 1.10
    from django.urls import ResolverMatch
except ImportError:
    # Django < 1.10
    from django.core.urlresolvers import ResolverMatch

from djblets.testing.testcases import TestCase


class PIISafePageURLTemplateTagTests(TestCase):
    """Unit tests for the {% pii_safe_page_url %} template tag."""

    def test_with_pii(self):
        """Testing {% pii_safe_page_url %} with PII in URL"""
        url = self._render_url('/users/test/', kwargs={
            'username': 'test',
        })

        self.assertEqual(url, '/users/&lt;REDACTED&gt;/')

    def test_without_pii(self):
        """Testing {% pii_safe_page_url %} without PII in URL"""
        self.assertEqual(self._render_url('/groups/test/'),
                         '/groups/test/')

    def _render_url(self, path, kwargs={}):
        request = RequestFactory().get(path)
        request.resolver_match = ResolverMatch(func=lambda: None,
                                               args=(),
                                               kwargs=kwargs)

        t = Template('{% load djblets_privacy %}'
                     '{% pii_safe_page_url %}')

        return t.render(Context({
            'request': request,
        }))



from __future__ import unicode_literals
import warnings

from djblets.db.managers import ConcurrencyManager
from djblets.deprecation import RemovedInDjblets20Warning


warnings.warn('djblets.util.db is deprecated', RemovedInDjblets20Warning)


__all__ = ['ConcurrencyManager']



#
# __init__.py -- Gravatar functionality
#
# Copyright (c) 2012       Beanbag, Inc.
# Copyright (c) 2008-2009  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from hashlib import md5
from warnings import warn

from django.conf import settings
from django.http import QueryDict
from django.utils import six

from djblets.deprecation import RemovedInDjblets20Warning


default_app_config = 'djblets.gravatars.apps.GravatarsAppConfig'


def get_gravatar_url_for_email(request=None, email=None, size=None):
    """Return the Gravatar URL for an e-mail address.

    The returned URL will always use HTTPS.

    Note that callers adhering to the GDPR should check for a user's consent
    before displaying a Gravatar on their behalf. This is checked automatically
    if using :py:mod:`djblets.avatars`.

    Args:
        request (django.http.HttpRequest):
            Ignored. This argument will be removed in Djblets 2.0.

        email (unicode):
            The e-mail address to get the Gravatar URL for.

        size (int, optional):
            An optional height and width of the image (in pixels).

    Returns:
        unicode:
        The URL for the Gravatar associated with the given e-mail address.
    """
    if request is not None:
        warn("djblets.gravatars.get_gravatar_url_for_email's request "
             "argument is deprecated and should be None. It will be removed "
             "in Djblets 2.0. Use get_gravatar_url_for_email(email=email, "
             "size=size) instead.",
             RemovedInDjblets20Warning)

    if email:
        email = email.strip().lower()

        if isinstance(email, six.text_type):
            email = email.encode('utf-8')

        email_hash = md5(email).hexdigest()
    else:
        email_hash = '00000000000000000000000000000000'

    url = 'https://secure.gravatar.com/avatar/%s' % email_hash
    params = QueryDict('', mutable=True)

    if not size and hasattr(settings, 'GRAVATAR_SIZE'):
        size = settings.GRAVATAR_SIZE

    if size:
        params['s'] = size

    if hasattr(settings, 'GRAVATAR_RATING'):
        params['r'] = settings.GRAVATAR_RATING

    if hasattr(settings, 'GRAVATAR_DEFAULT'):
        params['d'] = settings.GRAVATAR_DEFAULT

    if len(params):
        url = '%s?%s' % (url, params.urlencode())

    return url


def get_gravatar_url(request=None, user=None, size=None):
    """Return the Gravatar URL for a user.

    The returned URL will always use HTTPS.

    Note that callers adhering to the GDPR should check for a user's consent
    before displaying a Gravatar on their behalf. This is checked automatically
    if using :py:mod:`djblets.avatars`.

    Args:
        request (django.http.HttpRequest):
            Ignored. This argument will be removed in Djblets 2.0.

        user (django.contrib.auth.models.User):
            The user whose Gravatar URL is to be retrieved.

        size (int, optional):
            An optional height and width of the image (in pixels).

    Returns:
        unicode:
        The URL for the user's Gravatar.

    Raises:
        ValueError:
            Raised if ``user`` is ``None``.
    """
    if request is not None:
        warn("djblets.gravatars.get_gravatar_url's request request argument "
             "is deprecated and should be None. It will be removed in Djblets "
             "2.0. Use get_gravatar_url(user=user, size=size) instead.",
             RemovedInDjblets20Warning)

    if user is None:
        raise ValueError('"user" cannot be None.')

    if user.is_anonymous() or not user.email:
        return ''

    return get_gravatar_url_for_email(email=user.email, size=size)



"""Utility functions for working with template-related caches."""

from __future__ import unicode_literals

try:
    # Django >= 1.6, <= 1.8
    from django.template.base import libraries
except ImportError:
    libraries = None

try:
    # Django >= 1.7, <= 1.8
    from django.template.base import get_templatetags_modules
except ImportError:
    # Django >= 1.9
    get_templatetags_modules = None

try:
    # Django >= 1.8
    from django.template import Engine, engines
except ImportError:
    # Django >= 1.7, <= 1.8
    Engine = None
    engines = None


def clear_template_tag_caches():
    """Clear the template tags caches.

    This allows changes to the list of available template tags to be reflected
    in any new templates by emptying all the caches and forcing the available
    list of tags to be rebuilt.
    """
    if libraries is not None:
        # Django >= 1.6, <= 1.8
        libraries.clear()

    try:
        # Django == 1.6
        #
        # We'll import templatetags_modules here because we want the most
        # recent copy of templatetags_modules.
        from django.template.base import templatetags_modules

        del(templatetags_modules[:])
    except ImportError:
        if get_templatetags_modules is not None:
            # Django >= 1.7, <= 1.8
            get_templatetags_modules.cache_clear()

        if engines:
            # Django >= 1.8
            engines._engines.clear()
            Engine.get_default.cache_clear()


def clear_template_caches():
    """Clear the templates caches.

    This clears any caches for template parse trees and related state, forcing
    templates to be re-parsed and re-rendered.
    """
    if engines is not None:
        # Django >= 1.8
        template_loaders = []

        for engine in engines.all():
            template_loaders += engine.engine.template_loaders
    else:
        # Django >= 1.6, <= 1.7
        #
        # We need to import this here in order to get the latest copy
        # of this variable.
        from django.template.loader import template_source_loaders

        template_loaders = template_source_loaders or []

    for template_loader in template_loaders:
        try:
            template_loader.reset()
        except AttributeError:
            pass



from __future__ import unicode_literals

from kgb import SpyAgency

from django.conf import settings

from djblets.features import Feature, FeatureLevel, get_features_registry
from djblets.features.checkers import (BaseFeatureChecker, get_feature_checker,
                                       set_feature_checker)
from djblets.testing.testcases import TestCase


class DummyFeature(Feature):
    feature_id = 'djblets.tests.dummy-feature'
    name = 'Dummy Feature'
    summary = 'This is a dummy feature.'
    level = FeatureLevel.BETA

    def initialize(self):
        self._feature_inited = True

    def shutdown(self):
        self._feature_inited = False


class FeatureTests(SpyAgency, TestCase):
    """Unit tests for djblets.features.feature.Feature."""

    def setUp(self):
        super(FeatureTests, self).setUp()

        self._old_debug = settings.DEBUG

    def tearDown(self):
        super(FeatureTests, self).tearDown()

        settings.DEBUG = self._old_debug

        get_features_registry().reset()
        set_feature_checker(None)

    def test_init_registers(self):
        """Testing Feature.__init__ registers the feature"""
        features = get_features_registry()
        self.assertEqual(len(features), 0)

        feature = DummyFeature()

        self.assertEqual(len(features), 1)
        self.assertIn(feature, features)

    def test_init_calls_initialize(self):
        """Testing Feature.__init__ calls initialize"""
        feature = DummyFeature()

        self.assertTrue(getattr(feature, '_feature_inited', False))

    def test_is_enabled_with_stable(self):
        """Testing Feature.is_enabled with level=STABLE"""
        class StableFeature(DummyFeature):
            level = FeatureLevel.STABLE

        checker = get_feature_checker()

        feature = StableFeature()
        self.spy_on(checker.is_feature_enabled)

        self.assertTrue(feature.is_enabled())
        self.assertFalse(checker.is_feature_enabled.spy.called)

    def test_is_enabled_with_beta_debug_true(self):
        """Testing Feature.is_enabled with level=BETA and DEBUG=True"""
        class BetaFeature(DummyFeature):
            level = FeatureLevel.BETA

        feature = BetaFeature()

        settings.DEBUG = True
        self.assertTrue(feature.is_enabled())

    def test_is_enabled_with_beta_debug_false(self):
        """Testing Feature.is_enabled with level=BETA and DEBUG=False"""
        class BetaFeature(DummyFeature):
            level = FeatureLevel.BETA

        feature = BetaFeature()

        settings.DEBUG = False
        self.assertFalse(feature.is_enabled())

    def test_is_enabled_with_unavailable(self):
        """Testing Feature.is_enabled with level=UNAVAILABLE"""
        class UnavailableFeature(DummyFeature):
            level = FeatureLevel.UNAVAILABLE

        feature = UnavailableFeature()
        self.spy_on(get_feature_checker, call_original=False)

        self.assertFalse(feature.is_enabled())
        self.assertFalse(get_feature_checker.spy.called)

    def test_is_enabled_with_checker_true(self):
        """Testing Feature.is_enabled with feature checker returning True"""
        class DummyFeatureChecker(BaseFeatureChecker):
            def is_feature_enabled(self, feature_id, **kwargs):
                return True

        set_feature_checker(DummyFeatureChecker())

        feature = DummyFeature()
        self.assertTrue(feature.is_enabled())

    def test_is_enabled_with_checker_false(self):
        """Testing Feature.is_enabled with feature checker returning False"""
        class DummyFeatureChecker(BaseFeatureChecker):
            def is_feature_enabled(self, feature_id, **kwargs):
                return False

        set_feature_checker(DummyFeatureChecker())

        feature = DummyFeature()
        self.assertFalse(feature.is_enabled())



#
# views.py -- Views for the Admin UI.
#
# Copyright (c) 2010-2011  Beanbag, Inc.
# Copyright (c) 2008-2010  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import render
from django.template.context import RequestContext
from django.views.decorators.csrf import csrf_protect


@csrf_protect
@staff_member_required
def extension_list(request, extension_manager,
                   template_name='extensions/extension_list.html'):
    if request.method == 'POST':
        if 'full-reload' in request.POST:
            extension_manager.load(full_reload=True)

        return HttpResponseRedirect('.')
    else:
        # Refresh the extension list.
        extension_manager.load()

        return render(request, template_name)


@csrf_protect
@staff_member_required
def configure_extension(request, ext_class, form_class, extension_manager,
                        template_name='extensions/configure_extension.html'):
    extension = extension_manager.get_enabled_extension(ext_class.id)

    if not extension or not extension.is_configurable:
        raise Http404

    if request.method == 'POST':
        form = form_class(extension, request.POST, request.FILES)

        if form.is_valid():
            form.save()

            return HttpResponseRedirect(request.path + '?saved=1')
    else:
        form = form_class(extension)

    return render(request, template_name, {
        'extension': extension,
        'form': form,
        'saved': request.GET.get('saved', 0),
    })



from __future__ import unicode_literals

import warnings

import django
from django.conf.urls import include, url
from django.core.urlresolvers import (NoReverseMatch, clear_url_caches,
                                      reverse)
from django.utils import six
from django.views.decorators.cache import never_cache
from kgb import SpyAgency

from djblets.testing.testcases import TestCase
from djblets.urls.patterns import never_cache_patterns
from djblets.urls.resolvers import DynamicURLResolver


def dummy_view(request):
    pass


class URLPatternsTests(SpyAgency, TestCase):
    """Unit tests for djblets.urls.patterns."""

    def test_never_cache_patterns_with_prefix(self):
        """Testing never_cache_patterns with view lookup string prefix"""
        msg = (
            'String prefixes for URLs in never_cache_patterns() is '
            'deprecated, and will not work on Django 1.10 or higher.'
        )

        if django.VERSION[:2] >= (1, 10):
            with self.assertRaisesMessage(ValueError, msg):
                never_cache_patterns(
                    'djblets.urls.tests',

                    url('^a/$', dummy_view),
                )
        else:
            self.spy_on(never_cache)

            with warnings.catch_warnings(record=True) as w:
                urlpatterns = never_cache_patterns(
                    'djblets.urls.tests',

                    url('^a/$', 'dummy_view'),
                    url('^b/$', 'dummy_view'),
                )

            self.assertEqual(six.text_type(w[-1].message), msg)
            self.assertEqual(len(never_cache.spy.calls), 2)
            self.assertEqual(len(urlpatterns), 2)

            pattern = urlpatterns[0]
            self.assertEqual(pattern._callback_str,
                             'djblets.urls.tests.dummy_view')
            self.assertTrue(
                never_cache.spy.calls[0].returned(pattern.callback))

            pattern = urlpatterns[1]
            self.assertEqual(pattern._callback_str,
                             'djblets.urls.tests.dummy_view')
            self.assertTrue(
                never_cache.spy.calls[1].returned(pattern.callback))

    def test_never_cache_patterns_without_prefix(self):
        """Testing never_cache_patterns without view lookup string prefix"""
        self.spy_on(never_cache)

        with warnings.catch_warnings(record=True) as w:
            urlpatterns = never_cache_patterns(
                url('^a/$', dummy_view),
                url('^b/$', dummy_view),
            )

        self.assertEqual(len(w), 0)
        self.assertEqual(len(never_cache.spy.calls), 2)
        self.assertEqual(len(urlpatterns), 2)

        self.assertTrue(
            never_cache.spy.calls[0].returned(urlpatterns[0].callback))
        self.assertTrue(
            never_cache.spy.calls[1].returned(urlpatterns[1].callback))


class URLResolverTests(TestCase):
    def tearDown(self):
        super(URLResolverTests, self).tearDown()
        clear_url_caches()

    def test_dynamic_url_resolver(self):
        """Testing DynamicURLResolver"""
        def dummy_view(self):
            pass

        dynamic_urls = DynamicURLResolver()
        root_urlconf = [
            url(r'^root/', include([dynamic_urls])),
            url(r'^foo/', dummy_view, name='foo'),
        ]

        with self.settings(ROOT_URLCONF=root_urlconf):
            clear_url_caches()

            new_patterns = [
                url(r'^bar/$', dummy_view, name='bar'),
                url(r'^baz/$', dummy_view, name='baz'),
            ]

            # The new patterns shouldn't reverse, just the original "foo".
            reverse('foo')
            self.assertRaises(NoReverseMatch, reverse, 'bar')
            self.assertRaises(NoReverseMatch, reverse, 'baz')

            # Add the new patterns. Now reversing should work.
            dynamic_urls.add_patterns(new_patterns)

            reverse('foo')
            reverse('bar')
            reverse('baz')

            # Get rid of the patterns again. We should be back in the original
            # state.
            dynamic_urls.remove_patterns(new_patterns)

            reverse('foo')
            self.assertRaises(NoReverseMatch, reverse, 'bar')
            self.assertRaises(NoReverseMatch, reverse, 'baz')



from __future__ import unicode_literals

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import post_delete, post_save

from djblets.cache.synchronizer import GenerationSynchronizer
from djblets.integrations.errors import (IntegrationAlreadyRegisteredError,
                                         IntegrationNotRegisteredError,
                                         IntegrationRegistrationError)
from djblets.integrations.manager import (IntegrationManager,
                                          get_integration_managers,
                                          shutdown_integration_managers)
from djblets.integrations.integration import Integration
from djblets.integrations.tests.models import IntegrationConfig
from djblets.integrations.tests.testcases import IntegrationsTestCase


class BrokenIntegration(Integration):
    def __init__(self, *args, **kwargs):
        assert False


class DummyIntegration1(Integration):
    def initialize(self):
        pass


class DummyIntegration2(Integration):
    def initialize(self):
        pass


class IntegrationManagerTests(IntegrationsTestCase):
    """Unit tests for djblets.integrations.manager.IntegrationManager."""

    def setUp(self):
        super(IntegrationManagerTests, self).setUp()

        # Reset integration IDs, so we can check them appropriately in tests.
        DummyIntegration1.integration_id = None
        DummyIntegration2.integration_id = None

    def test_init_without_app(self):
        """Testing IntegrationManager initialization without INSTALLED_APPS"""
        msg = ('IntegrationManager requires djblets.integrations to be '
               'listed in settings.INSTALLED_APPS.')

        installed_apps = list(settings.INSTALLED_APPS)
        installed_apps.remove('djblets.integrations')

        with self.settings(INSTALLED_APPS=installed_apps):
            with self.assertRaisesMessage(ImproperlyConfigured, msg):
                IntegrationManager(IntegrationConfig)

    def test_init_without_middleware(self):
        """Testing IntegrationManager initialization without
        MIDDLEWARE_CLASSES
        """
        msg = ('IntegrationManager requires djblets.integrations.middleware.'
               'IntegrationsMiddleware to be listed in '
               'settings.MIDDLEWARE_CLASSES')

        middleware_classes = list(settings.MIDDLEWARE_CLASSES)
        middleware_classes.remove('djblets.integrations.middleware.'
                                  'IntegrationsMiddleware')

        with self.settings(MIDDLEWARE_CLASSES=middleware_classes):
            with self.assertRaisesMessage(ImproperlyConfigured, msg):
                IntegrationManager(IntegrationConfig)

    def test_shutdown(self):
        """Testing IntegrationManager.shutdown"""
        manager = IntegrationManager(IntegrationConfig)
        instance = manager.register_integration_class(DummyIntegration1)
        instance.enable_integration()

        self.assertTrue(instance.enabled)
        self.assertNotEqual(manager._integration_classes, {})
        self.assertNotEqual(manager._integration_instances, {})
        self.assertTrue(manager.is_expired())

        manager.shutdown()

        self.assertFalse(instance.enabled)
        self.assertEqual(manager._integration_classes, {})
        self.assertEqual(manager._integration_instances, {})
        self.assertFalse(manager.is_expired())

    def test_get_integration_classes(self):
        """Testing IntegrationManager.get_integration_classes"""
        manager = IntegrationManager(IntegrationConfig)
        manager.register_integration_class(DummyIntegration1)
        manager.register_integration_class(DummyIntegration2)

        classes = manager.get_integration_classes()
        self.assertEqual(set(classes), {DummyIntegration1, DummyIntegration2})

    def test_get_integration(self):
        """Testing IntegrationManager.get_integration"""
        manager = IntegrationManager(IntegrationConfig)
        manager.register_integration_class(DummyIntegration1)
        manager.register_integration_class(DummyIntegration2)

        integration = manager.get_integration(DummyIntegration1.integration_id)
        self.assertEqual(integration.__class__, DummyIntegration1)

        integration = manager.get_integration(DummyIntegration2.integration_id)
        self.assertEqual(integration.__class__, DummyIntegration2)

    def test_get_integration_with_invalid_id(self):
        """Testing IntegrationManager.get_integration with invalid ID"""
        manager = IntegrationManager(IntegrationConfig)

        with self.assertRaises(IntegrationNotRegisteredError):
            manager.get_integration('foobar')

    def test_get_integration_configs(self):
        """Testing IntegrationManager.get_integration_configs with all
        configurations
        """
        manager = IntegrationManager(IntegrationConfig)
        integration1 = manager.register_integration_class(DummyIntegration1)
        integration2 = manager.register_integration_class(DummyIntegration2)

        config1 = integration1.create_config(enabled=True, save=True)
        config2 = integration1.create_config(enabled=False, save=True)
        config3 = integration2.create_config(enabled=True, save=True)

        with self.assertNumQueries(1):
            configs = manager.get_integration_configs()

        self.assertEqual(set(configs), {config1, config2, config3})

    def test_get_integration_configs_with_class(self):
        """Testing IntegrationManager.get_integration_configs with instance
        class
        """
        manager = IntegrationManager(IntegrationConfig)
        integration1 = manager.register_integration_class(DummyIntegration1)
        integration2 = manager.register_integration_class(DummyIntegration2)

        config1 = integration1.create_config(enabled=True, save=True)
        config2 = integration1.create_config(enabled=False, save=True)
        integration2.create_config(enabled=True, save=True)

        with self.assertNumQueries(1):
            configs = manager.get_integration_configs(integration1)

        self.assertEqual(set(configs), {config1, config2})

    def test_get_integration_configs_with_filter(self):
        """Testing IntegrationManager.get_integration_configs with filter"""
        manager = IntegrationManager(IntegrationConfig)
        integration1 = manager.register_integration_class(DummyIntegration1)
        integration2 = manager.register_integration_class(DummyIntegration2)

        config1 = integration1.create_config(enabled=False, save=True)
        integration1.create_config(enabled=True, save=True)
        integration2.create_config(enabled=True, save=True)

        with self.assertNumQueries(1):
            configs = manager.get_integration_configs(integration1,
                                                      enabled=False)

        self.assertEqual(set(configs), {config1})

    def test_get_integration_configs_caches(self):
        """Testing IntegrationManager.get_integration_configs caches results"""
        manager = IntegrationManager(IntegrationConfig)
        integration1 = manager.register_integration_class(DummyIntegration1)

        config1 = integration1.create_config(enabled=True, save=True)
        config2 = integration1.create_config(enabled=False, save=True)

        with self.assertNumQueries(1):
            manager.get_integration_configs()

        with self.assertNumQueries(0):
            configs = manager.get_integration_configs()

        self.assertEqual(set(configs), {config1, config2})

    def test_clear_configs_cache(self):
        """Testing IntegrationManager.clear_configs_cache"""
        manager = IntegrationManager(IntegrationConfig)

        with self.assertNumQueries(1):
            manager.get_integration_configs()

        with self.assertNumQueries(0):
            manager.get_integration_configs()

        manager.clear_configs_cache()

        with self.assertNumQueries(1):
            manager.get_integration_configs()

    def test_clear_configs_cache_for_class(self):
        """Testing IntegrationManager.clear_configs_cache for class"""
        manager = IntegrationManager(IntegrationConfig)

        with self.assertNumQueries(1):
            manager.get_integration_configs(DummyIntegration1)

        with self.assertNumQueries(0):
            manager.get_integration_configs(DummyIntegration1)

        manager.clear_configs_cache(DummyIntegration1)

        with self.assertNumQueries(1):
            manager.get_integration_configs(DummyIntegration1)

    def test_clear_all_configs_cache(self):
        """Testing IntegrationManager.clear_all_configs_cache"""
        manager = IntegrationManager(IntegrationConfig)

        with self.assertNumQueries(1):
            manager.get_integration_configs(DummyIntegration1)

        with self.assertNumQueries(0):
            manager.get_integration_configs(DummyIntegration1)

        manager.clear_all_configs_cache()

        with self.assertNumQueries(1):
            manager.get_integration_configs(DummyIntegration1)

    def test_is_expired_new_instance(self):
        """Testing IntegrationManager.is_expired on a new instance"""
        manager = IntegrationManager(IntegrationConfig)

        self.assertFalse(manager.is_expired())

    def test_is_expired_after_registration(self):
        """Testing IntegrationManager.is_expired after new registration"""
        manager = IntegrationManager(IntegrationConfig)
        manager.register_integration_class(DummyIntegration1)

        self.assertTrue(manager.is_expired())

    def test_is_expired_after_config_saved(self):
        """Testing IntegrationManager.is_expired after config saved"""
        manager = IntegrationManager(IntegrationConfig)

        self.assertFalse(manager.is_expired())
        post_save.send(sender=IntegrationConfig)
        self.assertTrue(manager.is_expired())

    def test_is_expired_after_config_deleted(self):
        """Testing IntegrationManager.is_expired after config deleted"""
        manager = IntegrationManager(IntegrationConfig)

        self.assertFalse(manager.is_expired())
        post_delete.send(sender=IntegrationConfig)
        self.assertTrue(manager.is_expired())

    def test_is_expired_after_other_process_updates(self):
        """Testing IntegrationManager.is_expired after another process updates
        the configuration state
        """
        manager = IntegrationManager(IntegrationConfig)
        self.assertFalse(manager.is_expired())

        gen_sync = GenerationSynchronizer(manager._gen_sync.cache_key,
                                          normalize_cache_key=False)
        gen_sync.mark_updated()

        self.assertTrue(manager.is_expired())

    def test_check_expired_when_expired(self):
        """Testing IntegrationManager.check_expired when expired"""
        manager = IntegrationManager(IntegrationConfig)

        # Cache some state.
        integration1 = manager.register_integration_class(DummyIntegration1)
        integration1.create_config(enabled=True, save=True)
        manager.get_integration_configs(DummyIntegration1)

        integration2 = manager.register_integration_class(DummyIntegration2)
        manager.get_integration_configs(DummyIntegration1)

        # Make sure the integration isn't enabled. This would be because we
        # haven't refreshed yet.
        self.assertFalse(integration1.enabled)
        self.assertFalse(integration2.enabled)
        self.assertNotEqual(manager._integration_configs, {})

        # Check expired state.
        self.assertTrue(manager.is_expired())
        manager.check_expired()

        # Make sure state has been updated and caches cleared.
        self.assertFalse(manager.is_expired())
        self.assertTrue(integration1.enabled)
        self.assertFalse(integration2.enabled)
        self.assertEqual(manager._integration_configs, {})

    def test_check_expired_when_not_expired(self):
        """Testing IntegrationManager.check_expired when not expired"""
        manager = IntegrationManager(IntegrationConfig)

        # Cache some state.
        integration = manager.register_integration_class(DummyIntegration1)
        integration.create_config(enabled=True, save=True)
        manager.get_integration_configs(DummyIntegration1)

        # Make sure the integration isn't enabled. This would be because we
        # haven't refreshed yet.
        self.assertFalse(integration.enabled)
        self.assertNotEqual(manager._integration_configs, {})

        # Fake having the latest state.
        manager._needs_recalc = False

        # Check expired state, without actually expiring it.
        self.assertFalse(manager.is_expired())
        manager.check_expired()

        # Make sure state has not changed.
        self.assertFalse(manager.is_expired())
        self.assertFalse(integration.enabled)
        self.assertNotEqual(manager._integration_configs, {})

    def test_register_integration_class(self):
        """Testing IntegrationManager.register_integration_class"""
        manager = IntegrationManager(IntegrationConfig)

        integration = manager.register_integration_class(DummyIntegration1)
        self.assertEqual(integration.__class__, DummyIntegration1)

        integration_id = DummyIntegration1.integration_id
        self.assertEqual(
            integration_id,
            'djblets.integrations.tests.test_manager.DummyIntegration1')

        self.assertIn(integration_id, manager._integration_classes)
        self.assertEqual(manager._integration_classes[integration_id],
                         DummyIntegration1)
        self.assertIn(integration_id, manager._integration_instances)
        self.assertEqual(manager._integration_instances[integration_id],
                         integration)

    def test_register_integration_class_with_already_registered(self):
        """Testing IntegrationManager.register_integration_class with
        already registered class
        """
        manager = IntegrationManager(IntegrationConfig)
        manager.register_integration_class(DummyIntegration1)

        with self.assertRaises(IntegrationAlreadyRegisteredError):
            manager.register_integration_class(DummyIntegration1)

    def test_register_integration_class_with_construct_error(self):
        """Testing IntegrationManager.register_integration_class with
        construction error
        """
        manager = IntegrationManager(IntegrationConfig)

        with self.assertRaises(IntegrationRegistrationError):
            manager.register_integration_class(BrokenIntegration)

    def test_unregister_integration_class(self):
        """Testing IntegrationManager.unregister_integration_class"""
        manager = IntegrationManager(IntegrationConfig)
        manager.register_integration_class(DummyIntegration1)
        manager.unregister_integration_class(DummyIntegration1)

        integration_id = DummyIntegration1.integration_id

        self.assertNotIn(integration_id, manager._integration_classes)
        self.assertNotIn(integration_id, manager._integration_instances)

    def test_unregister_integration_class_with_enabled_integration(self):
        """Testing IntegrationManager.unregister_integration_class with
        enabled integration
        """
        manager = IntegrationManager(IntegrationConfig)
        integration = manager.register_integration_class(DummyIntegration1)

        integration_id = DummyIntegration1.integration_id
        integration.enabled = True

        manager.unregister_integration_class(DummyIntegration1)

        self.assertFalse(integration.enabled)
        self.assertNotIn(integration_id, manager._integration_classes)
        self.assertNotIn(integration_id, manager._integration_instances)

    def test_unregister_integration_class_with_unregistered(self):
        """Testing IntegrationManager.unregister_integration_class with
        unregistered integration
        """
        manager = IntegrationManager(IntegrationConfig)

        with self.assertRaises(IntegrationNotRegisteredError):
            manager.unregister_integration_class(DummyIntegration1)

    def test_get_integration_managers(self):
        """Testing get_integration_managers"""
        manager1 = IntegrationManager(IntegrationConfig)
        manager2 = IntegrationManager(IntegrationConfig)

        self.assertEqual(set(get_integration_managers()), {manager1, manager2})

    def test_shutdown_integration_managers(self):
        """Testing shutdown_integration_managers"""
        manager1 = IntegrationManager(IntegrationConfig)
        self.assertTrue(manager1.enabled)

        manager2 = IntegrationManager(IntegrationConfig)
        self.assertTrue(manager2.enabled)

        shutdown_integration_managers()

        self.assertEqual(get_integration_managers(), [])
        self.assertFalse(manager1.enabled)
        self.assertFalse(manager2.enabled)



"""Base support for configuration pages."""

from __future__ import unicode_literals

from django.template.context import RequestContext

from djblets.util.compat.django.template.loader import render_to_string


class ConfigPage(object):
    """Base class for a page of configuration forms.

    Each ConfigPage is represented in the main page by an entry in the
    navigation sidebar. When the user has navigated to that page, all visible
    :py:class:`djblets.configforms.forms.ConfigPageForm` subclasses owned by
    the ConfigPage will be displayed.
    """

    #: The unique ID of the page.
    #:
    #: This must be unique across all ConfigPages at a given URL.
    page_id = None

    #: The displayed title for the page.
    #:
    #: This will show up in the navigation sidebar.
    page_title = None

    #: The list of form subclasses to display on the page.
    form_classes = None

    #: The template used to render the page.
    template_name = 'configforms/config_page.html'

    def __init__(self, config_view, request, user):
        """Initialize the page.

        Args:
            config_view (ConfigPagesView):
                The view that manages this ConfigPage.

            request (HttpRequest):
                The HTTP request from the client.

            user (User):
                The user who is viewing the page.
        """
        self.config_view = config_view
        self.request = request
        self.forms = [
            form
            for form in (
                form_cls(self, request, user)
                for form_cls in self.form_classes
            )
            if form.is_visible()
        ]

    def is_visible(self):
        """Return whether the page should be visible.

        Visible pages are shown in the sidebar and can be navigated to.

        By default, a page is visible if at least one of its forms are
        also visible.

        Returns:
            bool:
                ``True`` if the page will be rendered, or ``False`` otherwise.
        """
        for form in self.forms:
            if form.is_visible():
                return True

        return False

    def render(self):
        """Render the page to a string.

        :py:attr:`template_name` will be used to render the page. The
        template will be passed ``page`` (this page's instance) and
        ``forms`` (the list of :py:class:`ConfigPageForm` instances to
        render).

        Subclasses can override this to provide additional rendering logic.

        Returns:
            unicode: The rendered page as HTML.
        """
        return render_to_string(
            template_name=self.template_name,
            context={
                'page': self,
                'forms': self.forms,
            },
            request=self.request)



from __future__ import unicode_literals

from django.conf import settings
from django.template import Context, Template, TemplateSyntaxError
from django.template.loader import get_template
from django.template.loaders.cached import Loader as CachedLoader

try:
    # Django >= 1.7
    from django.apps.registry import apps
except ImportError:
    # Django == 1.6
    apps = None

try:
    # Django >= 1.6, <= 1.8
    from django.template.base import get_templatetags_modules
except ImportError:
    # Django >= 1.9
    get_templatetags_modules = None

try:
    # Django >= 1.8
    from django.template import engines
except ImportError:
    # Django >= 1.7, <= 1.8
    engines = None

from djblets.template.caches import (clear_template_caches,
                                     clear_template_tag_caches)
from djblets.testing.testcases import TestCase


class CachesTests(TestCase):
    """Unit tests for djblets.template.caches."""

    def test_clear_template_tag_caches(self):
        """Testing clear_template_tag_caches"""

        def _check_state(enabled):
            if enabled:
                if get_templatetags_modules is not None:
                    self.assertIn(templatetags_module_name,
                                  get_templatetags_modules())

                self.assertEqual(
                    Template(template_str).render(Context({})),
                    'Hello, world!')
            else:
                if get_templatetags_modules is not None:
                    self.assertNotIn(templatetags_module_name,
                                     get_templatetags_modules())

                with self.assertRaisesRegexp(TemplateSyntaxError,
                                             'is not a (valid|registered) tag '
                                             'library'):
                    Template(template_str).render(Context({}))

        templatetags_module_name = 'djblets.template.tests.templatetags'
        template_str = (
            '{% load template_tests %}'
            '{% my_test_template_tag %}'
        )

        # Sanity-check that the template tag module isn't registered.
        _check_state(enabled=False)

        # Enable a new template tag library.
        old_installed_apps = settings.INSTALLED_APPS

        settings.INSTALLED_APPS = list(old_installed_apps) + [
            'djblets.template.tests',
        ]

        if apps:
            apps.set_installed_apps(settings.INSTALLED_APPS)

        try:
            clear_template_tag_caches()
            _check_state(enabled=True)
        finally:
            settings.INSTALLED_APPS = old_installed_apps

            if apps:
                apps.unset_installed_apps()

        clear_template_tag_caches()
        _check_state(enabled=False)

        # Other libraries should still work.
        Template('{% load djblets_js djblets_extensions %}').render(
            Context({}))

    def test_clear_template_caches(self):
        """Testing clear_template_caches"""
        # Load a template to populate the cache.
        get_template('avatars/avatar.html')

        if engines is not None:
            # Django >= 1.8
            template_loader = engines.all()[0].engine.template_loaders[0]
        else:
            # Django >= 1.6, <= 1.7
            #
            # We need to import this here in order to get the latest copy
            # of this variable.
            from django.template.loader import template_source_loaders

            template_loader = template_source_loaders[0]

        self.assertTrue(isinstance(template_loader, CachedLoader))

        if hasattr(template_loader, 'get_template'):
            self.assertNotEqual(template_loader.get_template_cache, {})
        else:
            self.assertNotEqual(template_loader.template_cache, {})

        clear_template_caches()

        if hasattr(template_loader, 'get_template'):
            self.assertEqual(template_loader.get_template_cache, {})
        else:
            self.assertEqual(template_loader.template_cache, {})



from __future__ import unicode_literals
import warnings

from djblets.cache.backend_compat import normalize_cache_backend
from djblets.deprecation import RemovedInDjblets20Warning


warnings.warn('djblets.util.cache is deprecated. Use '
              'djblets.cache.backend_compat.',
              RemovedInDjblets20Warning)


__all__ = ['normalize_cache_backend']



"""Built-in resource representing the Group model."""

from __future__ import unicode_literals

from django.contrib.auth.models import Group

from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.resources.registry import register_resource_for_model


class GroupResource(WebAPIResource):
    """A default resource for representing a Django Group model."""

    model = Group
    fields = ('id', 'name')

    uri_object_key = 'group_name'
    uri_object_key_regex = r'[A-Za-z0-9_\-]+'
    model_object_key = 'name'
    autogenerate_etags = True

    allowed_methods = ('GET',)


group_resource = GroupResource()
register_resource_for_model(Group, group_resource)



"""Djblets Avatar Services."""

from __future__ import unicode_literals

from djblets.avatars.services.base import AvatarService
from djblets.avatars.services.file_upload import FileUploadService
from djblets.avatars.services.gravatar import GravatarService
from djblets.avatars.services.url import URLAvatarService


__all__ = (
    'AvatarService',
    'FileUploadService',
    'GravatarService',
    'URLAvatarService',
)



#
# tests.py -- Unit tests for classes in djblets.feedview
#
# Copyright (c) 2008  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from django.test.utils import override_settings

from djblets.testing.testcases import TestCase


@override_settings(ROOT_URLCONF='djblets.feedview.test_urls')
class FeedViewTests(TestCase):
    def testViewFeedPage(self):
        """Testing view_feed with the feed-page.html template"""
        response = self.client.get('/feed/')
        self.assertContains(response, "Django 1.0 alpha released", 1)
        self.assertContains(response, "Introducing Review Board News", 1)

    def testViewFeedInline(self):
        """Testing view_feed with the feed-inline.html template"""
        response = self.client.get('/feed-inline/')
        self.assertContains(response, "Django 1.0 alpha released", 1)
        self.assertContains(response, "Introducing Review Board News", 1)

    def testViewFeedError(self):
        """Testing view_feed with a URL error"""
        response = self.client.get('/feed-error/')
        self.assertEqual(response.status_code, 200)
        self.assertTrue('error' in response.context)



"""Unit tests for djblets.configforms.forms.ConfigPageForm."""

from __future__ import unicode_literals

import warnings

from django import forms
from django.contrib.auth.models import User
from django.test.client import RequestFactory
from django.utils import six

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.deprecation import RemovedInDjblets20Warning
from djblets.testing.testcases import TestCase


class TestForm(ConfigPageForm):
    form_id = 'my-form'
    field1 = forms.CharField(label='Field 1',
                             required=False)
    field2 = forms.CharField(label='Field 2',
                             required=False)


class TestPage(ConfigPage):
    page_id = 'my-page'
    form_classes = [TestForm]


class ConfigPageFormTests(TestCase):
    """Unit tests for djblets.configforms.forms.ConfigPageForm."""

    def setUp(self):
        super(ConfigPageFormTests, self).setUp()

        request = RequestFactory().request()
        user = User.objects.create_user(username='test-user',
                                        password='test-user')
        page = TestPage(ConfigPagesView, request, user)

        self.form = TestForm(page, request, user)

    def test_initial_state(self):
        """Testing ConfigPageForm initial state"""
        self.assertEqual(self.form.fields['form_target'].initial, 'my-form')

    def test_profile(self):
        """Testing ConfigPageForm.profile raises a deprecation warning"""
        with warnings.catch_warnings(record=True) as w:
            try:
                self.form.profile
            except Exception:
                # This is either a SiteProfileNotAvailable (on Django 1.6) or
                # an AttributeError (Django 1.7+). Ignore it. We only care
                # about the warning.
                pass

        # We may get two warnings: One about the get_profile() deprecation in
        # Django 1.6, the other about our own deprecation. Ours will be first.
        message = w[0].message

        self.assertIsInstance(message, RemovedInDjblets20Warning)
        self.assertEqual(
            six.text_type(message),
            'ConfigFormPage.profile is deprecated. Update your code to '
            'fetch the profile manually instead.')

    def test_set_initial(self):
        """Testing ConfigPageForm.set_initial"""
        self.form.set_initial({
            'field1': 'foo',
            'field2': 'bar',
        })

        self.assertEqual(self.form.fields['field1'].initial, 'foo')
        self.assertEqual(self.form.fields['field2'].initial, 'bar')

    def test_render(self):
        """Testing ConfigPageForm.render"""
        rendered = self.form.render()

        self.assertHTMLEqual(
            '<input id="id_form_target" name="form_target"'
            ' type="hidden" value="my-form">'
            '<div class="fields-row field-field1" id="row-field1">'
            ' <div class="field">'
            '  <label for="id_field1">Field 1:</label>'
            '  <input id="id_field1" name="field1" type="text">'
            ' </div>'
            '</div>'
            '<div class="fields-row field-field2" id="row-field2">'
            ' <div class="field">'
            '  <label for="id_field2">Field 2:</label>'
            '  <input id="id_field2" name="field2" type="text">'
            ' </div>'
            '</div>'
            '<input type="submit" class="btn" value="Save">',
            rendered)



from __future__ import unicode_literals

try:
    from django.apps import AppConfig
except ImportError:
    # Django < 1.7
    AppConfig = object


class FeedViewAppConfig(AppConfig):
    name = 'djblets.feedview'
    label = 'djblets_feedview'



from __future__ import unicode_literals
import warnings

from djblets.db.evolution import FakeChangeFieldType
from djblets.deprecation import RemovedInDjblets20Warning


warnings.warn('djblets.util.dbevolution is deprecated. Use '
              'djblets.db.evolution instead.',
              RemovedInDjblets20Warning)


__all__ = ['FakeChangeFieldType']



from __future__ import unicode_literals

import os
from importlib import import_module

from django.template import TemplateDoesNotExist
from django.template.loaders import app_directories


class Loader(app_directories.Loader):
    """Looks for templates in app directories, optionally with a namespace.

    This extends the standard Django 'app_directories' template loader by
    allowing a prefix specifying the app whose template should be used.
    It solves the problem of one app defining a template and another app
    trying to both override and extend it, resulting in an infinite loop.

    Templates can be in the standard form of 'path/to/template', or in the
    namespaced form of 'app.path:path/to/template'.
    """
    def __init__(self, *args, **kwargs):
        super(Loader, self).__init__(*args, **kwargs)

        self._cache = {}

    def get_template_sources(self, template_name, template_dirs=None):
        parts = template_name.split(':')

        if len(parts) == 2:
            app = parts[0]

            template_dir = self._cache.get(app)

            if not template_dir:
                try:
                    mod = import_module(app)
                except ImportError:
                    raise TemplateDoesNotExist(template_name)

                template_dir = os.path.join(os.path.dirname(mod.__file__),
                                            'templates')
                self._cache[app] = template_dir

            template_name = parts[1]
            template_dirs = [template_dir]

        return super(Loader, self).get_template_sources(template_name,
                                                        template_dirs)



#!/usr/bin/env python

from __future__ import unicode_literals

import os
import sys

import django
from django.core.management import call_command

import djblets


if __name__ == '__main__':
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'djblets.settings')

    if hasattr(django, 'setup'):
        # Django >= 1.7
        django.setup()

    os.chdir(os.path.dirname(djblets.__file__))

    ret = call_command('compilemessages', interactive=False, verbosity=2)
    sys.exit(ret)



"""Unit tests for djblets.util.templatetags.djblets_forms."""

from __future__ import unicode_literals

from django.template import Context, Template
from pipeline.conf import settings as pipeline_settings

from djblets.testing.testcases import TestCase


class CompressedTagsTests(TestCase):
    """Unit tests for the {% compressed_* %} template tags."""

    def test_compressed_css_tag(self):
        """Testing {% compressed_css %}"""
        pipeline_settings.STYLESHEETS = {
            'test': {
                'source_filenames': [],
                'output_filename': 'test.css',
            }
        }

        t = Template('{% load compressed %}'
                     '{% compressed_css "test" %}')

        self.assertEqual(t.render(Context({'test': 'test'})),
                         '/test.css\n')

    def test_compressed_js_tag(self):
        """Testing {% compressed_js %}"""
        pipeline_settings.JAVASCRIPT = {
            'test': {
                'source_filenames': [],
                'output_filename': 'test.js',
            }
        }

        t = Template('{% load compressed %}'
                     '{% compressed_js "test" %}')

        self.assertEqual(t.render(Context({'test': 'test'})),
                         '/test.js\n')



"""Utility mixins for registries."""

from __future__ import unicode_literals

from djblets.registries.errors import ItemLookupError


class ExceptionFreeGetterMixin(object):
    """A mixin that prevents lookups from throwing errors."""

    def get(self, attr_name, attr_value):
        """Return the requested registered item.

        Args:
            attr_name (unicode):
                The attribute name.

            attr_value (object):
                The attribute value.

        Returns:
            object:
            The matching registered item, if found. Otherwise, ``None`` is
            returned.
        """
        try:
            return super(ExceptionFreeGetterMixin, self).get(attr_name,
                                                             attr_value)
        except ItemLookupError:
            return None



"""Feature switch support for applications.

This module contains convenience imports for:

* :py:class:`~djblets.features.feature.Feature`
* :py:class:`~djblets.features.feature.FeatureLevel`
* :py:class:`~djblets.features.registry.get_features_registry`
"""

from __future__ import unicode_literals

from djblets.features.feature import Feature, FeatureLevel
from djblets.features.registry import get_features_registry


__all__ = [
    'Feature',
    'FeatureLevel',
    'get_features_registry',
]



from __future__ import unicode_literals

from django.core.cache import cache

from djblets.cache.synchronizer import GenerationSynchronizer
from djblets.testing.testcases import TestCase


class GenerationSynchronizerTests(TestCase):
    """Unit tests for djblets.cache.synchronizer.GenerationSynchronizer."""

    def setUp(self):
        super(GenerationSynchronizerTests, self).setUp()

        self.gen_sync = GenerationSynchronizer('test-synchronizer')

    def test_initial_state(self):
        """Testing GenerationSynchronizer initial state"""
        self.assertIsNotNone(self.gen_sync.sync_gen)

    def test_is_expired_when_expired(self):
        """Testing GenerationSynchronizer.is_expired when expired"""
        cache.set(self.gen_sync.cache_key, self.gen_sync.sync_gen + 1)
        self.assertTrue(self.gen_sync.is_expired())

    def test_is_expired_when_not_expired(self):
        """Testing GenerationSynchronizer.is_expired when not expired"""
        self.assertFalse(self.gen_sync.is_expired())

    def test_refresh(self):
        """Testing GenerationSynchronizer.refresh"""
        new_sync_gen = self.gen_sync.sync_gen + 1
        cache.set(self.gen_sync.cache_key, new_sync_gen)

        self.gen_sync.refresh()
        self.assertEqual(self.gen_sync.sync_gen, new_sync_gen)

    def test_clear(self):
        """Testing GenerationSynchronizer.clear"""
        self.gen_sync.clear()
        self.assertEqual(cache.get(self.gen_sync.cache_key), None)

    def test_mark_updated(self):
        """Testing GenerationSynchronizer.mark_updated"""
        sync_gen = self.gen_sync.sync_gen

        self.gen_sync.mark_updated()
        self.assertEqual(self.gen_sync.sync_gen, sync_gen + 1)
        self.assertEqual(cache.get(self.gen_sync.cache_key),
                         self.gen_sync.sync_gen)



"""Markdown extension to safely escape HTML content.

This is built as a replacement for the ``safe_mode='escape'`` Markdown render
mode, which is deprecated in Python-Markdown 2.5.
"""

from __future__ import unicode_literals

from markdown.extensions import Extension


class EscapeHTMLExtension(Extension):
    """Extension to safely escape HTML content.

    This will remove any HTML preprocessors and patterns from the renderer,
    preventing HTML from being interpreted in any form. This results in the
    HTML being escaped instead, directly mimicking the behavior of the old
    ``safe_mode='escape'``.
    """

    def extendMarkdown(self, md, md_globals):
        """Extend the list of Markdown processors.

        Rather than extending, this will actually remove all HTML-based
        processors from the renderer, ensuring that the HTML is instead
        escaped.

        Args:
            md (markdown.Markdown):
                The Markdown renderer.

            md_globals (dict):
                Global variables from the :py:mod:`markdown` module.
        """
        # This is basically the implementation recommended by Python-Markdown.
        # In their implementation, these keys must exist. We make them
        # optional, so that if a caller is using both safe_mode='escape' and
        # this extension, it won't crash due to KeyErrors.
        md.preprocessors.pop('html_block', None)
        md.inlinePatterns.pop('html', None)


def makeExtension(*args, **kwargs):
    """Create and return an instance of this extension.

    Args:
        *args (tuple):
            Positional arguments for the extension.

        **kwargs (dict):
            Keyword arguments for the extension.
    """
    return EscapeHTMLExtension(*args, **kwargs)



from __future__ import unicode_literals

from django import forms
from django.utils.translation import ugettext_lazy as _

from djblets.forms.forms import KeyValueForm


class IntegrationConfigForm(KeyValueForm):
    """Base class for an integration settings form.

    This makes it easy to provide a basic form for manipulating the settings
    of an integration configuration. It takes care of loading/saving the
    values and prompting the user for a name.

    Integrations should subclass this and provide additional fields that they
    want to display to the user. They must provide a :py:class:`Meta` class
    containing the fieldsets they want to display.

    Applications can subclass this to provide additional special fields that
    should be made available to integrations. Integrations would then need
    to subclass that specialized form. Applications that want to do this will
    likely need to modify :py:attr:`model_fields`.
    """

    #: A list of fields on the model that should not be saved in settings.
    model_fields = ('name', 'enabled')

    #: The fieldset containing basic information on the configuration.
    #:
    #: Subclasses can override this to provide additional fields, styling, or
    #: a description.
    basic_info_fieldset = (None, {
        'fields': ('name', 'enabled'),
        'description': _(
            'Start by giving this configuration a name so you can easily '
            'identify it later. You can also mark this configuration as '
            'enabled or disabled.'
        ),
    })

    enabled = forms.BooleanField(
        label=_('Enable this integration'),
        initial=True,
        required=False)

    name = forms.CharField(
        label=_('Name'),
        required=True,
        widget=forms.widgets.TextInput(attrs={
            'size': 40,
        }))

    def __init__(self, integration, request, *args, **kwargs):
        """Initialize the form.

        Args:
            integration (djblets.integrations.integration.Integration):
                The integration being configured.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass to the form.

            **kwargs (dict):
                Keyword arguments to pass to the form.
        """
        self.integration = integration
        self.request = request

        self.build_fieldsets()

        super(IntegrationConfigForm, self).__init__(*args, **kwargs)

    @classmethod
    def build_fieldsets(cls):
        """Build the fieldsets used for the configuration form.

        By default, this will prepend :py:attr:`basic_info_fieldset` to
        the existing list of fieldsets on :py:class:`Meta` (if any).

        Subclasses can override this to provide more specialized
        customization of the form. Since this is working on a form class
        and not an instance, they should be careful to apply changes only
        once.
        """
        meta = getattr(cls, 'Meta', None)

        if meta:
            # Patch the list of fieldsets to contain the standard fieldset
            # for naming and enabling the configuration.
            fieldsets = getattr(meta, 'fieldsets', ())

            if not fieldsets or fieldsets[0] != cls.basic_info_fieldset:
                meta.fieldsets = (cls.basic_info_fieldset,) + fieldsets

    @property
    def config(self):
        """The configuration that's being edited.

        If this is a brand new configuration, this will be ``None`` until
        saved.

        Any value will be an instance of the subclass of
        :py:class:`~djblets.integrations.models.BaseIntegrationConfig` provided
        by the application supporting integrations.
        """
        return self.instance

    def get_key_value(self, key, default=None):
        """Return the value for a key.

        This will first look for the value from the
        :py:class:`~djblets.integrations.models.BaseIntegrationConfig`, and
        will then fall back to looking for the value in the configuration's
        :py:attr:`~djblets.integrations.models.BaseIntegrationConfig.settings`
        field.

        This is used internally by the parent class, and is not meant to be
        used directly.

        Args:
            key (unicode):
                The key in the configuration.

            default (object):
                The default value for the key, if not found.

        Returns:
            The value from the configuration.
        """
        if key in self.model_fields:
            return getattr(self.instance, key)
        else:
            return self.instance.get(key, default)

    def set_key_value(self, key, value):
        """Set the value for a key.

        This will first look for the key to set in the
        :py:class:`~djblets.integrations.models.BaseIntegrationConfig`. If
        it's there, that field's value will be set. Otherwise, it will fall
        back to setting the key in the configuration's
        :py:attr:`~djblets.integrations.models.BaseIntegrationConfig.settings`
        field.

        This is used internally by the parent class, and is not meant to be
        used directly.

        Args:
            key (unicode):
                The key in the configuration.

            value (object):
                The new value.
        """
        if key in self.model_fields:
            setattr(self.instance, key, value)
        else:
            self.instance.set(key, value)

    def create_instance(self):
        """Create an instance of a configuration.

        This is used internally by the parent class, and is not meant to be
        called by consumers of the form.

        Returns:
            djblets.integrations.models.BaseIntegrationConfig:
            A new integration configuration.
        """
        return self.integration.create_config()

    def save_instance(self):
        """Save the configuration.

        This is used internally by the parent class, and is not meant to be
        called by consumers of the form.
        """
        self.instance.save()



from __future__ import unicode_literals

from django.conf.urls import include, url


def build_integration_urlpatterns(list_view_cls, config_form_view_cls):
    """Build URL patterns for integration pages.

    This will produce a set of URL patterns for the integration administration
    pages that take advantage of the specialized views provided. These can
    be used in a :file:`urls.py`, either returned as a ``urlpatterns``
    variable or included using :py:func:`~django.conf.urls.include`.

    Args:
        list_view_cls (type):
            The view responsible for listing available integrations. This
            should be a subclass of
            :py:class:`~djblets.integrations.views.BaseIntegrationListView`.

        list_view_cls (type):
            The view responsible for configuring an integration. This should
            be a subclass of
            :py:class:`~djblets.integrations.views.BaseIntegrationConfigFormView`.
    """
    configs_urlpatterns = [
        url('^add/$', config_form_view_cls.as_view(),
            name='integration-add-config'),
        url('^(?P<config_id>[0-9]+)/$', config_form_view_cls.as_view(),
            name='integration-change-config'),
    ]

    return [
        url('^$', list_view_cls.as_view(), name='integration-list'),
        url('^(?P<integration_id>[A-Za-z0-9_\.]+)/configs/',
            include(configs_urlpatterns)),
    ]



from __future__ import unicode_literals

import json
import warnings

from django.core.exceptions import ValidationError
from django.db.models import Model
from django.utils import six

from djblets.db.fields import JSONField, JSONFormField
from djblets.deprecation import RemovedInDjblets20Warning
from djblets.testing.testcases import TestCase


class JSONFieldTests(TestCase):
    """Unit tests for JSONField."""

    def setUp(self):
        self.field = JSONField()

    def test_init_with_custom_encoder_class(self):
        """Testing JSONField initialization with custom encoder class"""
        class MyEncoder(json.JSONEncoder):
            def __init__(self, default_msg, **kwargs):
                self.default_msg = default_msg

                super(MyEncoder, self).__init__(**kwargs)

            def default(self, o):
                return self.default_msg

        with warnings.catch_warnings(record=True) as w:
            field = JSONField(
                encoder_cls=MyEncoder,
                encoder_kwargs={
                    'default_msg': 'What even is this?',
                })

        self.assertEqual(field.dumps(MyEncoder), '"What even is this?"')
        self.assertEqual(len(w), 0)

    def test_init_with_custom_encoder_instance(self):
        """Testing JSONField initialization with deprecated custom encoder
        instance
        """
        class MyEncoder(json.JSONEncoder):
            def default(self, o):
                return 'What even is this?'

        with warnings.catch_warnings(record=True) as w:
            field = JSONField(encoder=MyEncoder())

        self.assertEqual(field.dumps(MyEncoder), '"What even is this?"')
        self.assertEqual(len(w), 1)

        message = w[0].message
        self.assertIsInstance(message, RemovedInDjblets20Warning)
        self.assertEqual(six.text_type(message),
                         'The encoder argument to JSONField has been '
                         'replaced by the encoder_cls and encoder_kwargs '
                         'arguments. Support for encoder is deprecated.')

    def test_init_with_dict_value(self):
        """Testing JSONField initialization with initial dict value"""
        class MyModel(Model):
            myfield = JSONField()

        value = {
            'a': 1,
            'b': 2,
        }
        model = MyModel(myfield=value)

        # Make sure we're working with a copy in the model.
        value['c'] = 3

        self.assertEqual(model.myfield, {
            'a': 1,
            'b': 2,
        })
        self.assertEqual(model.get_myfield_json(),
                         '{"a": 1, "b": 2}')

    def test_init_with_dict_value_empty(self):
        """Testing JSONField initialization with initial empty dict value"""
        class MyModel(Model):
            myfield = JSONField()

        model = MyModel(myfield={})

        self.assertEqual(model.myfield, {})
        self.assertEqual(model.get_myfield_json(), '{}')

    def test_init_with_list_value(self):
        """Testing JSONField initialization with initial list value"""
        class MyModel(Model):
            myfield = JSONField()

        value = [1, 2]
        model = MyModel(myfield=value)

        # Make sure we're working with a copy in the model.
        value.append(3)

        self.assertEqual(model.myfield, [1, 2])
        self.assertEqual(model.get_myfield_json(), '[1, 2]')

    def test_init_with_list_value_empty(self):
        """Testing JSONField initialization with initial empty list value"""
        class MyModel(Model):
            myfield = JSONField()

        model = MyModel(myfield=[])

        self.assertEqual(model.myfield, [])
        self.assertEqual(model.get_myfield_json(), '[]')

    def test_init_with_json_string_value(self):
        """Testing JSONField initialization with initial JSON string value"""
        class MyModel(Model):
            myfield = JSONField()

        model = MyModel(myfield='{"a": 1, "b": 2}')

        self.assertEqual(model.myfield, {
            'a': 1,
            'b': 2,
        })
        self.assertEqual(model.get_myfield_json(),
                         '{"a": 1, "b": 2}')

    def test_init_with_json_string_value_empty(self):
        """Testing JSONField initialization with initial empty JSON string
        value
        """
        class MyModel(Model):
            myfield = JSONField()

        model = MyModel(myfield='')

        self.assertEqual(model.myfield, {})
        self.assertEqual(model.get_myfield_json(), '{}')

    def test_init_with_bad_type(self):
        """Testing JSONField initialization with initial unsupported value
        type
        """
        class MyModel(Model):
            myfield = JSONField()

        message = (
            "<class 'djblets.db.tests.test_json_field.MyModel'> is not a "
            "supported value type."
        )

        with self.assertRaisesMessage(ValidationError, message):
            MyModel(myfield=MyModel())

    def test_dumps_with_json_dict(self):
        """Testing JSONField with dumping a JSON dictionary"""
        result = self.field.dumps({'a': 1, 'b': 2})
        self.assertTrue(isinstance(result, six.string_types))
        self.assertEqual(result, '{"a": 1, "b": 2}')

    def test_dumps_with_json_string(self):
        """Testing JSONField with dumping a JSON string"""
        result = self.field.dumps('{"a": 1, "b": 2}')
        self.assertTrue(isinstance(result, six.string_types))
        self.assertEqual(result, '{"a": 1, "b": 2}')

    def test_loading_json_dict(self):
        """Testing JSONField with loading a JSON dictionary"""
        result = self.field.loads('{"a": 1, "b": 2}')
        self.assertTrue(isinstance(result, dict))
        self.assertTrue('a' in result)
        self.assertTrue('b' in result)

    def test_loading_json_broken_dict(self):
        """Testing JSONField with loading a badly serialized JSON dictionary"""
        result = self.field.loads('{u"a": 1, u"b": 2}')
        self.assertTrue(isinstance(result, dict))
        self.assertTrue('a' in result)
        self.assertTrue('b' in result)

    def test_loading_json_array(self):
        """Testing JSONField with loading a JSON array"""
        result = self.field.loads('[1, 2, 3]')
        self.assertTrue(isinstance(result, list))
        self.assertEqual(result, [1, 2, 3])

    def test_loading_string(self):
        """Testing JSONField with loading a stored string"""
        result = self.field.loads('"foo"')
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, {})

    def test_loading_broken_string(self):
        """Testing JSONField with loading a broken stored string"""
        result = self.field.loads('u"foo"')
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, {})

    def test_loading_python_code(self):
        """Testing JSONField with loading Python code"""
        result = self.field.loads('locals()')
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, {})

    def test_get_json(self):
        """Testing JSONField with get_{fieldname}_json"""
        class MyModel(Model):
            myfield = JSONField()

        model = MyModel()
        model.myfield = {
            'a': 1,
            'b': 2,
        }

        self.assertEqual(model.get_myfield_json(),
                         '{"a": 1, "b": 2}')

    def test_set_json(self):
        """Testing JSONField with set_{fieldname}_json"""
        class MyModel(Model):
            myfield = JSONField()

        model = MyModel()
        model.set_myfield_json('{"a": 1, "b": 2}')

        self.assertEqual(
            model.myfield,
            {
                'a': 1,
                'b': 2,
            })

    def test_validate_with_valid_json_string(self):
        """Testing JSONField with validating a valid JSON string"""
        self.field.run_validators('{"a": 1, "b": 2}')

    def test_validate_with_invalid_json_string(self):
        """Testing JSONField with validating an invalid JSON string"""
        self.assertRaises(ValidationError,
                          lambda: self.field.run_validators('foo'))

    def test_validate_with_json_dict(self):
        """Testing JSONField with validating a JSON dictionary"""
        self.field.run_validators({'a': 1, 'b': 2})


class JSONFormFieldTests(TestCase):
    """Unit tests for djblets.db.fields.JSONFormField."""

    def test_init_with_custom_encoder_class(self):
        """Testing JSONFormField initialization with custom encoder class"""
        class MyEncoder(json.JSONEncoder):
            def __init__(self, default_msg, **kwargs):
                self.default_msg = default_msg

                super(MyEncoder, self).__init__(**kwargs)

            def default(self, o):
                return self.default_msg

        field = JSONFormField(
            encoder_cls=MyEncoder,
            encoder_kwargs={
                'default_msg': 'What even is this?',
            })

        self.assertEqual(
            field.prepare_value({
                'a': 1,
                'b': 2,
                'cls': MyEncoder,
            }),
            '{\n'
            '  "a": 1,\n'
            '  "b": 2,\n'
            '  "cls": "What even is this?"\n'
            '}')

    def test_init_with_custom_encoder_instance(self):
        """Testing JSONFormField initialization with custom encoder instance"""
        class MyEncoder(json.JSONEncoder):
            def default(self, o):
                return 'What even is this?'

        field = JSONFormField(encoder=MyEncoder())

        self.assertEqual(
            field.prepare_value({
                'cls': MyEncoder,
            }),
            '{"cls": "What even is this?"}')

    def test_prepare_value_with_serialized(self):
        """Testing JSONFormField.prepare_value with serialized data"""
        field = JSONFormField()
        self.assertEqual(field.prepare_value('{"a": 1, "b": 2}'),
                         '{"a": 1, "b": 2}')

    def test_prepare_value_with_deserialized(self):
        """Testing JSONFormField.prepare_value with deserialized data"""
        field = JSONFormField()
        self.assertEqual(field.prepare_value({'a': 1, 'b': 2}),
                         '{\n'
                         '  "a": 1,\n'
                         '  "b": 2\n'
                         '}')

    def test_to_python_with_serialized(self):
        """Testing JSONFormField.to_python with serialized data"""
        field = JSONFormField()
        self.assertEqual(field.to_python('{"a": 1, "b": 2}'),
                         {'a': 1, 'b': 2})

    def test_to_python_with_deserialized(self):
        """Testing JSONFormField.to_python with deserialized data"""
        field = JSONFormField()
        self.assertEqual(field.to_python({"a": 1, "b": 2}),
                         {'a': 1, 'b': 2})

    def test_to_python_with_empty_string(self):
        """Testing JSONFormField.to_python with empty string"""
        field = JSONFormField()
        self.assertIsNone(field.to_python(''))

    def test_to_python_with_empty_non_string(self):
        """Testing JSONFormField.to_python with empty non-string value"""
        field = JSONFormField()
        self.assertEqual(field.to_python({}), {})
        self.assertEqual(field.to_python([]), [])

    def test_to_python_with_validation_error(self):
        """Testing JSONFormField.to_python with bad JSON data triggering
        ValidationError
        """
        field = JSONFormField()

        with self.assertRaises(ValidationError) as cm:
            field.to_python('{a: 1}')

        self.assertEqual(cm.exception.message,
                         'Expecting property name enclosed in double quotes: '
                         'line 1 column 2 (char 1)')
        self.assertEqual(cm.exception.code, 'invalid')
        self.assertEqual(cm.exception.params, {
            'value': '{a: 1}',
        })



"""Tests for WebAPI signals."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from kgb import SpyAgency

from djblets.testing.testcases import TestCase, TestModelsLoaderMixin
from djblets.webapi.models import BaseWebAPIToken
from djblets.webapi.signals import webapi_token_created, webapi_token_updated
from djblets.webapi.tests.test_api_token import WebAPIToken


class WebAPITokenSignalsTests(SpyAgency, TestModelsLoaderMixin, TestCase):
    """Tests for WebAPI token signals."""

    tests_app = 'djblets.webapi.tests'

    def setUp(self):
        super(WebAPITokenSignalsTests, self).setUp()

        self.user = User.objects.create_user(username='test_user',
                                             email='test@example.com')

    def test_webapi_token_created(self):
        """Testing WebAPIToken.objects.generate_token() emits
        webapi_token_created
        """
        def on_webapi_token_created(auto_generated, **kwargs):
            self.assertFalse(auto_generated)

        try:
            webapi_token_created.connect(on_webapi_token_created)
            self.spy_on(on_webapi_token_created)
            WebAPIToken.objects.generate_token(user=self.user)
            self.assertTrue(on_webapi_token_created.spy.called)
        finally:
            webapi_token_created.disconnect(on_webapi_token_created)

    def test_webapi_token_created_autogenerated(self):
        """Testing WebAPIToken.objects.generate_token() with
        auto_generated=True emits webapi_token_created with
        auto_generated=True
        """
        def on_webapi_token_created(auto_generated, **kwargs):
            self.assertTrue(auto_generated)

        try:
            webapi_token_created.connect(on_webapi_token_created)
            self.spy_on(on_webapi_token_created)
            WebAPIToken.objects.generate_token(user=self.user,
                                               auto_generated=True)
            self.assertTrue(on_webapi_token_created.spy.called)
        finally:
            webapi_token_created.disconnect(on_webapi_token_created)

    def test_webapi_token_updated(self):
        """Testing WebAPIToken.save() emits webapi_token_updated"""
        def on_webapi_token_created(**kwargs):
            pass

        def on_webapi_token_updated(**kwargs):
            pass

        try:
            webapi_token_created.connect(on_webapi_token_created)
            webapi_token_updated.connect(on_webapi_token_updated)

            self.spy_on(on_webapi_token_created)
            self.spy_on(on_webapi_token_updated)

            token = WebAPIToken.objects.generate_token(user=self.user)

            self.assertTrue(on_webapi_token_created.spy.called)
            self.assertFalse(on_webapi_token_updated.spy.called)

            on_webapi_token_created.spy.reset_calls()

            token.save()

            self.assertFalse(on_webapi_token_created.spy.called)
            self.assertTrue(on_webapi_token_updated.spy.called)
        finally:
            webapi_token_created.disconnect(on_webapi_token_created)
            webapi_token_updated.disconnect(on_webapi_token_updated)



"""Unit tests for djblets.configforms.pages.ConfigPage."""

from __future__ import unicode_literals

import re

from django.contrib.auth.models import User
from django.test.client import RequestFactory

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.testing.testcases import TestCase


class TestForm1(ConfigPageForm):
    form_id = 'my-form-1'
    form_title = 'Form 1'


class TestForm2(ConfigPageForm):
    form_id = 'my-form-2'
    form_title = 'Form 2'


class TestForm3(ConfigPageForm):
    form_id = 'my-form-3'
    form_title = 'Form 3'

    def is_visible(self):
        return False


class TestPage(ConfigPage):
    page_id = 'my-page'
    form_classes = [TestForm1, TestForm2, TestForm3]


class ConfigPageTests(TestCase):
    """Unit tests for djblets.configforms.pages.ConfigPage."""

    def setUp(self):
        super(ConfigPageTests, self).setUp()

        self.request = RequestFactory().request()
        self.user = User.objects.create_user(username='test-user',
                                             password='test-user')
        self.page = TestPage(ConfigPagesView, self.request, self.user)

    def test_initial_state(self):
        """Testing ConfigPage initial state"""
        self.assertEqual(len(self.page.forms), 2)
        self.assertIsInstance(self.page.forms[0], TestForm1)
        self.assertIsInstance(self.page.forms[1], TestForm2)

    def test_is_visible_with_visible_forms(self):
        """Testing ConfigPage.is_visible with visible forms"""
        self.assertTrue(self.page.is_visible())

    def test_is_visible_with_no_visible_forms(self):
        """Testing ConfigPage.is_visible without visible forms"""
        class TestPage(ConfigPage):
            page_id = 'my-page'
            form_classes = [TestForm3]

        page = TestPage(ConfigPagesView, self.request, self.user)
        self.assertFalse(page.is_visible())

    def test_render(self):
        """Testing ConfigPage.render"""
        # Filter out the CSRF token, since it's hard to match.
        rendered = re.sub(r"<input.+name='csrfmiddlewaretoken'.*>",
                          '',
                          self.page.render())

        self.assertHTMLEqual(
            rendered,
            '<div class="box-container">'
            ' <div class="box">'
            '  <div class="box-inner">'
            '   <div class="box-head">'
            '    <h1 class="box-title">Form 1</h1>'
            '   </div>'
            '   <div class="box-main box-foot">'
            '    <form method="post" action=".#my-page" id="form_my-form-1">'
            '     <input id="id_form_target" name="form_target"'
            '            type="hidden" value="my-form-1">'
            '     <input type="submit" class="btn" value="Save">'
            '    </form>'
            '   </div>'
            '  </div>'
            ' </div>'
            '</div>'
            '<div class="box-container">'
            ' <div class="box">'
            '  <div class="box-inner">'
            '   <div class="box-head">'
            '    <h1 class="box-title">Form 2</h1>'
            '   </div>'
            '   <div class="box-main box-foot">'
            '    <form method="post" action=".#my-page" id="form_my-form-2">'
            '     <input id="id_form_target" name="form_target"'
            '            type="hidden" value="my-form-2">'
            '     <input type="submit" class="btn" value="Save">'
            '    </form>'
            '   </div>'
            '  </div>'
            ' </div>'
            '</div>')



"""Test case support for full API testing.

This provides a useful mixin, :py:class:`WebAPITestCaseMixin`, that can be
mixed into a base test case for an API testing suite. This mixin provides
a number of useful helpers for calling APIs and validating results, along
with generating consistent docstrings.
"""

from __future__ import print_function, unicode_literals

import json
import pprint
import warnings

from django.test.client import BOUNDARY, MULTIPART_CONTENT, encode_multipart
from django.utils import six

from djblets.deprecation import RemovedInDjblets20Warning
from djblets.testing.testcases import TestCase


class WebAPITestCaseMixin(TestCase):
    """Mixin for test cases for thoroughly testing APIs.

    This helps with the testing of APIs by providing convenient methods for
    performing HTTP requests and checking their response payloads, redirects,
    mimetypes, and more.

    Any test methods in a subclass that have the ``is_test_template`` flag set
    to ``True`` can include ``<URL>`` in their docstring. When the test is
    run, the ``<URL>`` will be replaced by the contents of the subclass's
    :py:attr:`sample_api_url`.

    WebAPITestCaseMixin is meant to be mixed into a new base class that all
    API test suites would inherit from. Subclasses must provide values for
    :py:attr:`sample_api_url` and :py:attr:`error_mimetype`.
    """

    #: The path to the resource, which may be placed into the docstring.
    sample_api_url = None

    #: The mimetype representing an error for this API.
    error_mimetype = None

    #: A fake base URL for API operations.
    base_url = 'http://testserver'

    def shortDescription(self):
        """Return a description for a particular test.

        If the test has the ``is_test_template`` attribute set, and
        contains ``<URL>`` in the docstring, the URL will be replaced by
        the contents of :py:attr:`sample_api_url`.

        Returns:
            unicode:
            The description of the test.
        """
        desc = super(WebAPITestCaseMixin, self).shortDescription()

        if self.sample_api_url:
            test_method = getattr(self, self._testMethodName)

            if getattr(test_method, 'is_test_template', False):
                desc = desc.replace('<URL>', self.sample_api_url)

        return desc

    def api_get(self, *args, **kwargs):
        """Perform and check a HTTP GET request, returning additional data.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        This method is a wrapper around :py:meth:`api_get_with_response`. See
        that method for the arguments.

        Args:
            *args (tuple):
                Positional arguments to pass to
                :py:meth:`api_get_with_response`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`api_get_with_response`.

        Returns:
            dict or bytes:
            The parsed payload data as a dictionary, if ``expected_json`` is
            ``True`` and the response isn't a HTTP 302. Otherwise, the
            raw payload contents.
        """
        return self.api_get_with_response(*args, **kwargs)[0]

    def api_get_with_response(self, path, data={}, **kwargs):
        """Perform and check a HTTP GET request to an API resource.

        This will perform the request to the resource and validate that all the
        results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            data (object, optional):
                The request query string.

            *args (tuple):
                Positional arguments to pass to :py:meth:`api_call`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`api_call`.

        Returns:
            tuple:
            A 2-item tuple containing the parsed response and the original
            :py:class:`HttpResponse`.
        """
        return self.api_call(client_http_method=self.client.get,
                             path=path,
                             data=data,
                             **kwargs)

    def api_post(self, *args, **kwargs):
        """Perform and check an HTTP POST to an API resource.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        This method is a wrapper around :py:meth:`api_post_with_response`. See
        that method for the arguments.

        Args:
            *args (tuple):
                Positional arguments to pass to
                :py:meth:`api_post_with_response`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`api_post_with_response`.

        Returns:
            dict or bytes:
            The parsed payload data as a dictionary, if ``expected_json`` is
            ``True`` and the response isn't a HTTP 302. Otherwise, the
            raw payload contents.
        """
        return self.api_post_with_response(*args, **kwargs)[0]

    def api_post_with_response(self, path, data={}, expected_status=201,
                               **kwargs):
        """Perform an HTTP POST to an API resource, returning additional data.

        This works like :py:meth:`api_post`, but returns the resulting payload
        and the original :py:class:`HttpResponse`.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            data (object, optional):
                The request payload.

            expected_status (int, optional):
                The expected HTTP status.

            *args (tuple):
                Positional arguments to pass to :py:meth:`api_call`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`api_call`.

        Returns:
            tuple:
            A 2-item tuple containing the parsed response and the original
            :py:class:`HttpResponse`.
        """
        return self.api_call(client_http_method=self.client.post,
                             path=path,
                             data=data,
                             expected_status=expected_status,
                             **kwargs)

    def api_put(self, *args, **kwargs):
        """Perform and check an HTTP PUT to an API resource.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        This method is a wrapper around :py:meth:`api_put_with_response`. See
        that method for the arguments.

        Args:
            *args (tuple):
                Positional arguments to pass to
                :py:meth:`api_put_with_response`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`api_put_with_response`.

        Returns:
            dict or bytes:
            The parsed payload data as a dictionary, if ``expected_json`` is
            ``True`` and the response isn't a HTTP 302. Otherwise, the
            raw payload contents.
        """
        return self.api_put_with_response(*args, **kwargs)[0]

    def api_put_with_response(self, path, data={},
                              content_type=MULTIPART_CONTENT,
                              **kwargs):
        """Perform an HTTP PUT to an API resource, returning additional data.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            data (object, optional):
                The request payload.

            *args (tuple):
                Positional arguments to pass to :py:meth:`api_call`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`api_call`.

        Returns:
            A 2-item tuple containing the parsed response and the original
            :py:class:`HttpResponse`.
        """
        return self.api_call(client_http_method=self.client.put,
                             path=path,
                             data=data,
                             content_type=content_type,
                             **kwargs)

    def api_delete(self, *args, **kwargs):
        """Perform and check an HTTP DELETE to an API resource.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        This method is a wrapper around :py:meth:`api_delete_with_response`.
        See that method for the arguments.

        Args:
            *args (tuple):
                Positional arguments to pass to
                :py:meth:`api_delete_with_response`.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`api_delete_with_response`.

        Returns:
            bytes:
            The HTTP response payload. This won't have any content if the
            delete was successful.
        """
        return self.api_delete_with_response(*args, **kwargs)[0]

    def api_delete_with_response(self, path, expected_status=204, **kwargs):
        """Perform an HTTP DELETE to an API resource, returning additional data.

        This will perform the request to the resource and validate that all
        the results are what the caller expects.

        Args:
            path (unicode):
                The path to the resource to request.

            expected_status (int, optional):
                The expected HTTP status.

            *args (tuple):
                Positional arguments to pass to :py:meth:`api_call`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`api_call`.

        Returns:
            tuple:
            A 2-item tuple containing the parsed response and the original
            :py:class:`HttpResponse`.
        """
        return self.api_call(client_http_method=self.client.delete,
                             path=path,
                             expected_status=expected_status,
                             **kwargs)

    def assertHttpOK(self, response, check_last_modified=False,
                     check_etag=False):
        """Assert that a response was OK and optionally has caching headers.

        Args:
            response (django.http.HttpResponse):
                The HTTP response from the API.

            check_last_modified (bool):
                If set, this will assert that the response has a
                ``Last-Modified`` header.

            check_etag (bool):
                If set, this will assert that the response has a
                ``ETag`` header.
        """
        self.assertEquals(response.status_code, 200)

        if check_last_modified:
            self.assertIn('Last-Modified', response)

        if check_etag:
            self.assertIn('ETag', response)

    def assertHttpNotModified(self, response):
        """Assert that the response was HTTP 304 Not Modified.

        Args:
            response (django.http.HttpResponse):
                The HTTP response from the API.
        """
        self.assertEquals(response.status_code, 304)
        self.assertEquals(response.content, '')

    def api_call(self, client_http_method, path, data=None,
                 follow_redirects=False, expected_status=200,
                 expected_redirects=[], expected_headers={},
                 expected_mimetype=None, expected_num_queries=None,
                 expected_json=True, return_http_response=True, **extra):
        """Perform an API call using a client API method.

        This will invoke an API function with all the necessary parameters,
        and check the results for the expected values.

        Args:
            api_func (callable):
                The API function to call.

            path (unicode):
                The path to the resource to request.

            data (dict, optional):
                The data to pass in the request. For an HTTP GET, this will
                be used as a query string. For other requests, this will be
                the request body.

            follow_redirects (bool, optional):
                Whether to expect and follow redirects to another URL.

            expected_status (int, optional):
                The expected HTTP status.

            expected_redirects (list of unicode, optional):
                The list of expected redirects performed by the resource(s),
                in order.

            expected_headers (dict, optional):
                Expected HTTP headers and their values from the
                response.

            expected_num_queries (int, optional):
                The number of queries this API call is expected to make.

            expected_mimetype (unicode, optional):
                The expected mimetype for the response payload.

            expected_json (bool):
                Whether JSON-parsable content is expected in the response.

            return_http_response (bool, optional):
                Whether to return the :py:class:`~django.http.HttpResponse`
                as part of the result.

            **extra (dict):
                Extra data to pass to the client HTTP method.

        Returns:
            object:
            By default, this returns the payload content, which may be a
            raw byte string result or a deserialized JSON body (depending on
            ``expected_json``).

            If passing ``return_http_response=True``, this will return a
            tuple in the form of ``(payload_content, http_response)``,
            where ``http_response`` is the
            :py:class:`~django.http.HttpResponse` object.
        """
        def _call_api():
            return client_http_method(path=path,
                                      data=data,
                                      follow=follow_redirects,
                                      HTTP_X_REQUESTED_WITH='XMLHttpRequest',
                                      **extra)

        # Normalize the API path so that the base URL containing the hostname
        # is stripped.
        if path.startswith(self.base_url):
            path = path[len(self.base_url):]

        # The `query` argument is outdated. If passed, move its value over to
        # `data` and raise a deprecation warning.
        if 'query' in extra:
            warnings.warn('%s should be passed a `data` argument. `query` '
                          'is deprecated and will be removed in Djblets 2.0.',
                          RemovedInDjblets20Warning)

            data = extra.pop('query')

        # If the caller is explicitly requested multipart content, ensure we've
        # encoded the data.
        if extra.get('content_type') == MULTIPART_CONTENT:
            data = encode_multipart(BOUNDARY, data)

        # Log some details about the API request that's about to be performed.
        print('Performing HTTP %s for API %s'
              % (client_http_method.__name__.upper(), path))

        if data is not None:
            print('Request data = %r' % data)

        if expected_num_queries is None:
            response = _call_api()
        else:
            with self.assertNumQueries(expected_num_queries):
                response = _call_api()

        print('Raw API response: %r' % response.content)

        rsp = response.content

        self.assertEqual(response.status_code, expected_status)

        if expected_status in (204, 405):
            self.assertEqual(response.content, '')
            rsp = None
        else:
            if expected_status != 302 and expected_json:
                rsp = json.loads(response.content)
            else:
                rsp = response.content

            print('Parsed API response:')
            pprint.pprint(rsp)

            if expected_status >= 400:
                # Error responses should be using the test's error mimetype
                # and not some valid response mimetype.
                self.assertIsNone(expected_mimetype)

                if expected_status != 405:
                    self.assertEqual(response['Content-Type'],
                                     self.error_mimetype)
            elif expected_status != 302:
                # All status codes other than the few above should have a
                # response payload matching the expected mimetype.
                self.assertIsNotNone(expected_mimetype)
                self.assertEqual(response['Content-Type'], expected_mimetype)

        # Check if the response redirected the way the caller expected.
        if expected_redirects:
            self.assertEqual(len(response.redirect_chain),
                             len(expected_redirects))

            for redirect in expected_redirects:
                self.assertEqual(response.redirect_chain[0][0],
                                 self.base_url + expected_redirects[0])

        # Check that all the expected headers are present in the response.
        for header, value in six.iteritems(expected_headers):
            self.assertIn(header, response)
            self.assertEqual(response[header], value)

        if return_http_response:
            return rsp, response
        else:
            return rsp
